# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

from cpymad.madx import Madx

import xpart as xp
import xobjects as xo
import xtrack as xt
import xfields as xf

###############################
# Enable pyheadtail interface #
###############################

xt.enable_pyheadtail_interface()

############
# Settings #
############

gpu_device = 0
seq_name = "sps"
qx0,qy0 = 20.13, 20.18
p0c = 26e9
cavity_name = "actcse.31632"
cavity_lag = 180
frequency = 200e6
rf_voltage = 4e6
use_wakes = True
n_slices_wakes = 100
limit_z = 0.7
bunch_intensity = 1e11
sigma_z = 22.5e-2
nemitt_x=2.5e-6
nemitt_y=2.5e-6
n_part=int(1e4)
num_turns=2
num_spacecharge_interactions = 540

# mode = 'frozen'
mode = 'pic'

#######################################################
# Use cpymad to load the sequence and match the tunes #
#######################################################

mad = Madx()
mad.call("./madx_sps/sps.seq")
mad.call("./madx_sps/lhc_q20.str")
mad.call("./madx_sps/macro.madx")
mad.beam()
mad.use(seq_name)

mad.twiss()
tw_thick = mad.table.twiss.dframe()
summ_thick = mad.table.summ.dframe()

mad.input("""select, flag=makethin, slice=1, thick=false;
            makethin, sequence=sps, style=teapot, makedipedge=false;""")
mad.use(seq_name)
mad.exec(f"sps_match_tunes({qx0},{qy0});")

##################
# Switch context #
##################

if gpu_device is None:
   context = xo.ContextCpu()
else:
   context = xo.ContextCupy(device=gpu_device)

#################################
# Build line from MAD-X lattice #
#################################

line = xt.Line.from_madx_sequence(sequence=mad.sequence[seq_name],
           deferred_expressions=True, install_apertures=True,
           apply_madx_errors=False)
# Define reference particle
line.particle_ref = xt.Particles(p0c=p0c,mass0=xt.PROTON_MASS_EV)

################
# Switch on RF #
################

line[cavity_name].voltage = rf_voltage
line[cavity_name].lag = cavity_lag
line[cavity_name].frequency = frequency

############################################
# Twiss (check that the lattice is stable) #
############################################

line.build_tracker()
tw = line.twiss()

# We unfreeze the line so that we can still insert elements in the
# original one (would be prevented by the existence of a tracker linked to the
# line).
line.unfreeze()

#####################################
# Install spacecharge interactions) #
#####################################

# Install frozen space-charge lenses

lprofile = xf.LongitudinalProfileQGaussian(
       number_of_particles=bunch_intensity,
       sigma_z=sigma_z,
       z0=0.,
       q_parameter=1.)

xf.install_spacecharge_frozen(line=line,
                  particle_ref=line.particle_ref,
                  longitudinal_profile=lprofile,
                  nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                  sigma_z=sigma_z,
                  num_spacecharge_interactions=num_spacecharge_interactions,
                  )

# Switch to PIC or quasi-frozen
if mode == 'frozen':
   pass # Already configured in line
elif mode == 'quasi-frozen':
   xf.replace_spacecharge_with_quasi_frozen(
                                   line,
                                   update_mean_x_on_track=True,
                                   update_mean_y_on_track=True)
elif mode == 'pic':
   pic_collection, all_pics = xf.replace_spacecharge_with_PIC(
       _context=context, line=line,
       n_sigmas_range_pic_x=8,
       n_sigmas_range_pic_y=8,
       nx_grid=256, ny_grid=256, nz_grid=100,
       n_lims_x=7, n_lims_y=3,
       z_range=(-3*sigma_z, 3*sigma_z))
else:
   raise ValueError(f'Invalid mode: {mode}')

######################
# Install wakefields #
######################

if use_wakes:

   # We rescale the waketable to take into account for the difference between
   # the average beta functions (for which the table is calculated) and the beta
   # functions at the location in the lattice at which the wake element is
   # installed.
   wakefields = np.genfromtxt('wakes/kickerSPSwake_2020_oldMKP.wake')
   wakefields[:,1] *= 54.65/tw['betx'][0]
   wakefields[:,2] *= 54.51/tw['bety'][0]
   wakefields[:,3] *= 54.65/tw['betx'][0]
   wakefields[:,4] *= 54.65/tw['betx'][0]
   wakefile = 'wakes/wakefields.wake'
   np.savetxt(wakefile,wakefields,delimiter='\t')

   # Build PyHEADTAIL wakefield element
   from PyHEADTAIL.particles.slicing import UniformBinSlicer
   from PyHEADTAIL.impedances.wakes import WakeTable, WakeField
   n_slices_wakes = 500
   slicer_for_wakefields = UniformBinSlicer(n_slices_wakes, z_cuts=(-limit_z, limit_z))
   waketable = WakeTable(wakefile, ['time', 'dipole_x', 'dipole_y',
                  'quadrupole_x', 'quadrupole_y'])
   wakefield = WakeField(slicer_for_wakefields, waketable)

   # Specity that the wakefield element needs to run on CPU and that lost
   # particles need to be hidden for this element (required by PyHEADTAIL)
   wakefield.needs_cpu = True
   wakefield.needs_hidden_lost_particles = True

   # Insert element in the line
   line.insert_element(element=wakefield,name="wakefield", at_s=0)

#################
# Build Tracker #
#################

line.build_tracker(_context=context)

######################
# Generate particles #
######################

# (we choose to match the distribution without accounting for spacecharge)
line_sc_off = line.filter_elements(exclude_types_starting_with='SpaceCh')

particles = xp.generate_matched_gaussian_bunch(_context=context,
        num_particles=n_part, total_intensity_particles=bunch_intensity,
        nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
        particle_ref=line.particle_ref, line=line_sc_off)
particles.circumference = line.get_length() # Needed by pyheadtail

###########################################################
# We use a phase monitor to measure the tune turn by turn #
###########################################################

phasem = xp.PhaseMonitor(line,
                 num_particles=n_part, twiss=line_sc_off.twiss())

#########
# Track #
#########

for turn in range(num_turns):
   phasem.measure(particles)
   #import pdb; pdb.set_trace()
   line.track(particles)

##################
# Plot footprint #
##################

import matplotlib.pyplot as plt
plt.close('all')
f,ax = plt.subplots()
ax.plot(phasem.qx, phasem.qy,'b.',ms=1)
ax.set_xlim(0, 0.5)
ax.set_ylim(0, 0.5)
plt.show()



import xtrack as xt

import numpy as np

magnet = xt.Magnet(k0=0.002, h=0.002, k1=0.02, length=2)

m_exact = magnet.copy()
m_exact.model = 'bend-kick-bend'
m_exact.integrator='yoshida4'
m_exact.num_multipole_kicks = 1000

m_expanded = magnet.copy()
m_expanded.model = 'mat-kick-mat'
m_expanded.integrator='yoshida4'
m_expanded.num_multipole_kicks = 1000

p0 = xt.Particles(x=1e-3, y=2e-3)
px_list = [1e-7, 2e-7, 5e-7, 1e-6, 2e-6, 5e-6,
           1e-5, 2e-5, 5e-5, 1e-4, 2e-4, 5e-4, 1e-3, 2e-3, 5e-3, 1e-2]
px_in = np.array(px_list)

x_out_exact = []
px_out_exact = []
y_out_exact = []
py_out_exact = []
zeta_out_exact = []

x_out_expanded = []
px_out_expanded = []
y_out_expanded = []
py_out_expanded = []
zeta_out_expanded = []

for dd in px_list:
    p_exact = p0.copy()
    p_exact.px = dd
    m_exact.track(p_exact)

    p_expanded = p0.copy()
    p_expanded.px = dd
    m_expanded.track(p_expanded)

    x_out_exact.append(p_exact.x[0])
    px_out_exact.append(p_exact.px[0])
    y_out_exact.append(p_exact.y[0])
    py_out_exact.append(p_exact.py[0])
    zeta_out_exact.append(p_exact.zeta[0])

    x_out_expanded.append(p_expanded.x[0])
    px_out_expanded.append(p_expanded.px[0])
    y_out_expanded.append(p_expanded.y[0])
    py_out_expanded.append(p_expanded.py[0])
    zeta_out_expanded.append(p_expanded.zeta[0])


# Cast to numpy arrays
import numpy as np
x_out_exact = np.array(x_out_exact)
px_out_exact = np.array(px_out_exact)
y_out_exact = np.array(y_out_exact)
py_out_exact = np.array(py_out_exact)
zeta_out_exact = np.array(zeta_out_exact)

x_out_expanded = np.array(x_out_expanded)
px_out_expanded = np.array(px_out_expanded)
y_out_expanded = np.array(y_out_expanded)
py_out_expanded = np.array(py_out_expanded)
zeta_out_expanded = np.array(zeta_out_expanded)


import matplotlib.pyplot as plt

plt.close('all')
plt.figure(1, figsize=(12, 8))

# Plot for px
ax0 = plt.subplot(2, 3, 1)
plt.loglog(px_list, np.abs(px_out_exact - px_out_expanded), '.-')
plt.ylabel('Error in px out')
plt.xlabel('px in')

# Plot for x
plt.subplot(2, 3, 2, sharex=ax0, sharey=ax0)
plt.loglog(px_list, np.abs(x_out_exact - x_out_expanded), '.-')
plt.ylabel('Error in x out')
plt.xlabel('px in')

# Plot for py
plt.subplot(2, 3, 3, sharex=ax0, sharey=ax0)
plt.loglog(px_list, np.abs(py_out_exact - py_out_expanded), '.-')
plt.ylabel('Error in py out')
plt.xlabel('px in')


# Plot for y
plt.subplot(2, 3, 4, sharex=ax0, sharey=ax0)
plt.loglog(px_list, np.abs(y_out_exact - y_out_expanded), '.-')
plt.ylabel('Error in y out')
plt.xlabel('px in')

# Plot for zeta
plt.subplot(2, 3, 5, sharex=ax0, sharey=ax0)
plt.loglog(px_list, np.abs(zeta_out_exact - zeta_out_expanded), '.-')
plt.ylabel('Error in zeta out')
plt.xlabel('px in')


plt.subplots_adjust(wspace=0.3, hspace=0.3)


plt.show()








# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import pathlib

import numpy as np
import pytest
from cpymad.madx import Madx

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

path = test_data_folder.joinpath('hllhc14_input_mad/')


@pytest.fixture(scope='module')
def mad_with_errors():
    mad_with_errors = Madx(stdout=False)
    mad_with_errors.call(str(path.joinpath("final_seq.madx")))
    mad_with_errors.use(sequence='lhcb1')
    mad_with_errors.twiss()
    mad_with_errors.readtable(file=str(path.joinpath("final_errors.tfs")),
                              table="errtab")
    mad_with_errors.seterr(table="errtab")
    mad_with_errors.set(format=".15g")

    mad_with_errors.sequence.lhcb1.beam.sigt = 1e-10
    mad_with_errors.sequence.lhcb1.beam.sige = 1e-10
    mad_with_errors.sequence.lhcb1.beam.et = 1e-10

    return mad_with_errors


@pytest.fixture(scope='module')
def mad_b12_no_errors():
    mad_b12_no_errors = Madx(stdout=False)
    mad_b12_no_errors.call(str(test_data_folder.joinpath(
                                   'hllhc15_noerrors_nobb/sequence.madx')))
    mad_b12_no_errors.globals['vrf400'] = 16
    mad_b12_no_errors.globals['lagrf400.b1'] = 0.5
    mad_b12_no_errors.globals['lagrf400.b2'] = 0
    mad_b12_no_errors.use(sequence='lhcb1')
    mad_b12_no_errors.twiss()
    mad_b12_no_errors.use(sequence='lhcb2')
    mad_b12_no_errors.twiss()

    return mad_b12_no_errors


@pytest.fixture(scope='module')
def mad_b4_no_errors():
    mad_b4_no_errors = Madx(stdout=False)
    mad_b4_no_errors.call(str(test_data_folder.joinpath(
                                   'hllhc15_noerrors_nobb/sequence_b4.madx')))
    mad_b4_no_errors.globals['vrf400'] = 16
    mad_b4_no_errors.globals['lagrf400.b2'] = 0
    mad_b4_no_errors.use(sequence='lhcb2')
    mad_b4_no_errors.twiss()

    return mad_b4_no_errors


surv_starting_point = {
    "theta0": -np.pi / 9, "psi0": np.pi / 7, "phi0": np.pi / 11,
    "X0": -300, "Y0": 150, "Z0": -100}


b4_b2_mapping = {
    'mbxf.4l1..1': 'mbxf.4l1..4',
    'mbxf.4l5..1': 'mbxf.4l5..4',
    'mb.b19r5.b2..1': 'mb.b19r5.b2..2',
    'mb.b19r1.b2..1': 'mb.b19r1.b2..2',
    }


@pytest.mark.parametrize('configuration', ['b1_with_errors', 'b2_no_errors'])
@for_all_test_contexts
def test_twiss_and_survey(
        test_context,
        mad_with_errors,
        mad_b4_no_errors,
        mad_b12_no_errors,
        configuration,
):
    if configuration == 'b1_with_errors':
        mad_load = mad_with_errors
        mad_ref = mad_with_errors
        ref_element_for_mu = 'acsca.d5l4.b1'
        seq_name = 'lhcb1'
        reverse = False
        use = False
        range_for_partial_twiss = ('mb.b19r3.b1..1', 'mb.b19l3.b1..1')
    elif configuration == 'b2_no_errors':
        mad_load = mad_b4_no_errors
        mad_ref = mad_b12_no_errors
        ref_element_for_mu = 'acsca.d5r4.b2'
        seq_name = 'lhcb2'
        reverse = True
        use = True
        range_for_partial_twiss = ('mb.b19l3.b2..1', 'mb.b19r3.b2..1')

    if use:
        mad_ref.use(sequence=seq_name)
        mad_load.use(sequence=seq_name)

    # I want only the betatron part in the sigma matrix
    mad_ref.sequence[seq_name].beam.sigt = 1e-10
    mad_ref.sequence[seq_name].beam.sige = 1e-10
    mad_ref.sequence[seq_name].beam.et = 1e-10

    # I set asymmetric emittances
    mad_ref.sequence[seq_name].beam.exn = 2.5e-6
    mad_ref.sequence[seq_name].beam.eyn = 3.5e-6

    twmad = mad_ref.twiss(chrom=True)
    survmad = mad_ref.survey(**surv_starting_point)

    line_full = xt.Line.from_madx_sequence(
            mad_load.sequence[seq_name], apply_madx_errors=True)
    line_full.elements[10].iscollective = True # we make it artificially collective to test this option
    line_full.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1,
                            gamma0=mad_load.sequence[seq_name].beam.gamma)

    # Test twiss also on simplified line
    line_simplified = line_full.copy()

    print('Simplifying line...')
    line_simplified.remove_inactive_multipoles()
    line_simplified.merge_consecutive_multipoles()
    line_simplified.remove_zero_length_drifts()
    line_simplified.merge_consecutive_drifts()
    line_simplified.use_simple_bends()
    line_simplified.use_simple_quadrupoles()
    print('Done simplifying line')

    line_full.line = None
    line_simplified.tracker = None

    line_full.build_tracker(_context=test_context)
    assert line_full.iscollective

    line_simplified.build_tracker(_context=test_context)

    for simplified, line in zip((False, True), [line_full, line_simplified]):

        print(f"Simplified: {simplified}")

        twxt = line.twiss()
        twxt4d = line.twiss(method='4d')
        survxt = line.survey(**surv_starting_point)

        if reverse:
            twxt = twxt.reverse()
            twxt4d = twxt4d.reverse()
            survxt = survxt.reverse()

        assert len(twxt.name) == len(line.element_names) + 1

        # Check value_at_element_exit (not implemented for now, to be reintroduced)
        # if not reverse: # TODO: to be generalized...
        #     twxt_exit = line.twiss(values_at_element_exit=True)
        #     for nn in['s', 'x','px','y','py', 'zeta','delta','ptau',
        #             'betx','bety','alfx','alfy','gamx','gamy','dx','dpx','dy',
        #             'dpy','mux','muy', 'name']:
        #         assert np.all(twxt[nn][1:] == twxt_exit[nn])

        # Twiss a part of the machine
        tw_init = line.twiss().get_twiss_init(
                                    at_element=range_for_partial_twiss[0])
        tw4d_init = line.twiss(method='4d').get_twiss_init(
                                    at_element=range_for_partial_twiss[0])
        tw_part = line.twiss(start=range_for_partial_twiss[0],
                end=range_for_partial_twiss[1], init=tw_init)
        tw4d_part = line.twiss(method='4d',
                start=range_for_partial_twiss[0],
                end=range_for_partial_twiss[1], init=tw4d_init)
        if reverse:
            tw_part = tw_part.reverse()
            tw4d_part = tw4d_part.reverse()

        ipart_start = line.element_names.index(range_for_partial_twiss[0])
        ipart_stop = line.element_names.index(range_for_partial_twiss[1])
        assert len(tw_part.name) == ipart_stop - ipart_start + 2
        assert len(tw4d_part.name) == ipart_stop - ipart_start + 2

        # Check against mad
        for twtst in [twxt, twxt4d]:
            xo.assert_allclose(mad_ref.table.summ.q1[0], twtst['qx'], rtol=1e-4, atol=0)
            xo.assert_allclose(mad_ref.table.summ.q2[0], twtst['qy'], rtol=1e-4, atol=0)
            xo.assert_allclose(mad_ref.table.summ.dq1, twtst['dqx'], atol=0.1, rtol=0)
            xo.assert_allclose(mad_ref.table.summ.dq2, twtst['dqy'], atol=0.1, rtol=0)
            xo.assert_allclose(mad_ref.table.summ.alfa[0],
                twtst['momentum_compaction_factor'], atol=7e-8, rtol=0)
            if twtst is not tw4d_part:
                xo.assert_allclose(twxt['qs'], 0.0021, atol=1e-4, rtol=0)


        for is_part, twtst in zip([False, False, True, True],
                                   [twxt, twxt4d, tw_part, tw4d_part]):

            nemitt_x = mad_ref.sequence[seq_name].beam.exn
            nemitt_y = mad_ref.sequence[seq_name].beam.eyn
            Sigmas = twtst.get_betatron_sigmas(nemitt_x, nemitt_y)

            for nn in twtst._col_names:
                assert len(twtst[nn]) == len(twtst['name'])

            test_at_elements = []
            test_at_elements.extend(['mbxf.4l1..1', 'mbxf.4l5..1'])

            if seq_name.endswith('b1'):
                test_at_elements.extend(['mb.b19r5.b1..1', 'mb.b19r1.b1..1'])
            elif seq_name.endswith('b2'):
                test_at_elements.extend(['mb.b19r5.b2..1', 'mb.b19r1.b2..1'])

            if line is line_full:
                test_at_elements += ['ip1', 'ip2', 'ip5', 'ip8']

            for name in test_at_elements:

                if reverse:
                    name_mad = b4_b2_mapping.get(name, name)
                else:
                    name_mad = name

                imad = list(twmad['name']).index(name_mad+':1')
                ixt = list(twtst['name']).index(name) + 1 # MAD measures at exit

                eemad = mad_ref.sequence[seq_name].expanded_elements[name]

                mad_shift_x = eemad.align_errors.dx if eemad.align_errors else 0
                mad_shift_y = eemad.align_errors.dy if eemad.align_errors else 0

                xo.assert_allclose(twtst['s'][ixt], twmad['s'][imad],
                                atol=1e-6, rtol=0)
                xo.assert_allclose(twtst['betx'][ixt], twmad['betx'][imad],
                                atol=0, rtol=7e-4)
                xo.assert_allclose(twtst['bety'][ixt], twmad['bety'][imad],
                                atol=0, rtol=7e-4)
                xo.assert_allclose(twtst['alfx'][ixt], twmad['alfx'][imad],
                                atol=1e-1, rtol=0)
                xo.assert_allclose(twtst['alfy'][ixt], twmad['alfy'][imad],
                                atol=1e-1, rtol=0)
                xo.assert_allclose(twtst['dx'][ixt], twmad['dx'][imad],
                                atol=1e-2, rtol=0)
                xo.assert_allclose(twtst['dy'][ixt], twmad['dy'][imad],
                                atol=1e-2, rtol=0)
                xo.assert_allclose(twtst['dpx'][ixt], twmad['dpx'][imad],
                                atol=3e-4, rtol=0)
                xo.assert_allclose(twtst['dpy'][ixt], twmad['dpy'][imad],
                                atol=3e-4, rtol=0)

                if is_part:
                    # I chck the phase advance w.r.t. ip1
                    mux0_mad = twmad['mux'][list(twmad.name).index(ref_element_for_mu + ':1')]
                    muy0_mad = twmad['muy'][list(twmad.name).index(ref_element_for_mu + ':1')]
                    mux0_tst = twtst['mux'][list(twtst.name).index(ref_element_for_mu)]
                    muy0_tst = twtst['muy'][list(twtst.name).index(ref_element_for_mu)]
                else:
                    # I check the absolute phase advance
                    mux0_mad = 0
                    muy0_mad = 0
                    mux0_tst = 0
                    muy0_tst = 0
                xo.assert_allclose(twtst['mux'][ixt] - mux0_tst,
                                  twmad['mux'][imad] - mux0_mad,
                                  atol=1e-4, rtol=0)
                xo.assert_allclose(twtst['muy'][ixt] - muy0_tst,
                                  twmad['muy'][imad] - muy0_mad,
                                  atol=1e-4, rtol=0)

                xo.assert_allclose(twtst['s'][ixt], twmad['s'][imad],
                                atol=5e-6, rtol=0)

                # I check the orbit relative to sigma to be more accurate at the IP
                sigx = np.sqrt(twmad['sig11'][imad])
                sigy = np.sqrt(twmad['sig33'][imad])

                xo.assert_allclose(twtst['x'][ixt], twmad['x'][imad],
                                atol=0.03*sigx, rtol=0)
                xo.assert_allclose(twtst['y'][ixt],
                                (twmad['y'][imad]),
                                atol=0.03*sigy, rtol=0)

                xo.assert_allclose(twtst['px'][ixt], twmad['px'][imad],
                                atol=2e-7, rtol=0)
                xo.assert_allclose(twtst['py'][ixt], twmad['py'][imad],
                                atol=2e-7, rtol=0)

                xo.assert_allclose(Sigmas.Sigma11[ixt], twmad['sig11'][imad], atol=5e-10)
                xo.assert_allclose(Sigmas.Sigma12[ixt], twmad['sig12'][imad], atol=3e-12)
                xo.assert_allclose(Sigmas.Sigma22[ixt], twmad['sig22'][imad], atol=1e-12)
                xo.assert_allclose(Sigmas.Sigma33[ixt], twmad['sig33'][imad], atol=5e-10)
                xo.assert_allclose(Sigmas.Sigma34[ixt], twmad['sig34'][imad], rtol=1e-4, atol=3e-12)
                xo.assert_allclose(Sigmas.Sigma44[ixt], twmad['sig44'][imad], atol=1e-12)

                if twtst not in [twxt4d, tw4d_part]: # 4d less precise due to different momentum (coupling comes from feeddown)
                    xo.assert_allclose(Sigmas.Sigma13[ixt], twmad['sig13'][imad], atol=5e-10)
                    xo.assert_allclose(Sigmas.Sigma14[ixt], twmad['sig14'][imad], atol=2e-12)
                    xo.assert_allclose(Sigmas.Sigma23[ixt], twmad['sig23'][imad], atol=1e-12)
                    xo.assert_allclose(Sigmas.Sigma24[ixt], twmad['sig24'][imad], atol=1e-12)

                # check matrix is symmetric
                xo.assert_allclose(Sigmas.Sigma12[ixt], Sigmas.Sigma21[ixt], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma13[ixt], Sigmas.Sigma31[ixt], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma14[ixt], Sigmas.Sigma41[ixt], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma23[ixt], Sigmas.Sigma32[ixt], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma24[ixt], Sigmas.Sigma42[ixt], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma34[ixt], Sigmas.Sigma43[ixt], atol=1e-16)

                # check matrix consistency with Sigma.Sigma
                xo.assert_allclose(Sigmas.Sigma11[ixt], Sigmas.Sigma[ixt][0, 0], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma12[ixt], Sigmas.Sigma[ixt][0, 1], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma13[ixt], Sigmas.Sigma[ixt][0, 2], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma14[ixt], Sigmas.Sigma[ixt][0, 3], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma21[ixt], Sigmas.Sigma[ixt][1, 0], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma22[ixt], Sigmas.Sigma[ixt][1, 1], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma23[ixt], Sigmas.Sigma[ixt][1, 2], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma24[ixt], Sigmas.Sigma[ixt][1, 3], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma31[ixt], Sigmas.Sigma[ixt][2, 0], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma32[ixt], Sigmas.Sigma[ixt][2, 1], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma33[ixt], Sigmas.Sigma[ixt][2, 2], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma34[ixt], Sigmas.Sigma[ixt][2, 3], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma41[ixt], Sigmas.Sigma[ixt][3, 0], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma42[ixt], Sigmas.Sigma[ixt][3, 1], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma43[ixt], Sigmas.Sigma[ixt][3, 2], atol=1e-16)
                xo.assert_allclose(Sigmas.Sigma44[ixt], Sigmas.Sigma[ixt][3, 3], atol=1e-16)

                # Check sigma_x, sigma_y
                xo.assert_allclose(Sigmas.sigma_x[ixt], np.sqrt(Sigmas.Sigma11[ixt]), atol=1e-16)
                xo.assert_allclose(Sigmas.sigma_y[ixt], np.sqrt(Sigmas.Sigma33[ixt]), atol=1e-16)
                xo.assert_allclose(Sigmas.sigma_px[ixt], np.sqrt(Sigmas.Sigma22[ixt]), atol=1e-16)
                xo.assert_allclose(Sigmas.sigma_py[ixt], np.sqrt(Sigmas.Sigma44[ixt]), atol=1e-16)

                if not(is_part): # We don't have survey on a part of the machine
                    # Check survey
                    xo.assert_allclose(survxt.X[ixt], survmad['X'][imad], atol=1e-6)
                    xo.assert_allclose(survxt.Y[ixt], survmad['Y'][imad], rtol=2e-7, atol=1e-6)
                    xo.assert_allclose(survxt.Z[ixt], survmad['Z'][imad], atol=1e-6)
                    xo.assert_allclose(survxt.s[ixt], survmad['s'][imad], atol=5e-6)
                    xo.assert_allclose(survxt.phi[ixt], survmad['phi'][imad], atol=1e-10)
                    xo.assert_allclose(survxt.theta[ixt], survmad['theta'][imad], atol=1e-10)
                    xo.assert_allclose(survxt.psi[ixt], survmad['psi'][imad], atol=1e-10)

                    # angle and tilt are associated to the element itself (ixt - 1)
                    # For now not checking the sign of the angles, convetion in mad-X to be calrified
                    xo.assert_allclose(np.abs(survxt.angle[ixt-1]),
                            np.abs(survmad['angle'][imad]), atol=1e-10)
                    xo.assert_allclose(survxt.tilt[ixt-1], survmad['tilt'][imad], atol=1e-10)

        # Check to_pandas (not extensively for now)
        dftw = twtst.to_pandas()
        dfsurv = survxt.to_pandas()
        assert np.all(dftw['s'] == twtst['s'])
        assert np.all(dfsurv['s'] == survxt['s'])

        # Test custom s locations
        if not reverse:
            s_test = [2e3, 1e3, 3e3, 10e3]
            twats = line.twiss(at_s = s_test)
            for ii, ss in enumerate(s_test):
                xo.assert_allclose(twats['s'][ii], ss, rtol=0, atol=1e-14)
                xo.assert_allclose(twats['alfx'][ii], np.interp(ss, twxt['s'], twxt['alfx']),
                                rtol=1e-5, atol=0)
                xo.assert_allclose(twats['alfy'][ii], np.interp(ss, twxt['s'], twxt['alfy']),
                                rtol=1e-5, atol=0)
                xo.assert_allclose(twats['dpx'][ii], np.interp(ss, twxt['s'], twxt['dpx']),
                                rtol=1e-5, atol=0)
                xo.assert_allclose(twats['dpy'][ii], np.interp(ss, twxt['s'], twxt['dpy']),
                                rtol=1e-5, atol=0)


def norm(x):
    return np.sqrt(np.sum(np.array(x) ** 2))


@for_all_test_contexts
def test_line_import_from_madx(test_context, mad_with_errors):

    mad = mad_with_errors

    rtol = 1e-7
    atol = 1e-14

    print('Build line with expressions...')
    line_with_expressions = xt.Line.from_madx_sequence(
        mad.sequence['lhcb1'],
        apply_madx_errors=True,
        install_apertures=True,
        deferred_expressions=True,
    )
    line_with_expressions.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV,
                        q0=1, gamma0=mad.sequence.lhcb1.beam.gamma)

    print('Build line without expressions...')
    line_no_expressions = xt.Line.from_madx_sequence(
        mad.sequence['lhcb1'],
        apply_madx_errors=True,
        install_apertures=True,
        deferred_expressions=False,
    )
    line_no_expressions.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV,
                        q0=1, gamma0=mad.sequence.lhcb1.beam.gamma)

    # Profit to test to_dict/from_dict
    print('Test to_dict/from_dict')
    line_with_expressions = xt.Line.from_dict(line_with_expressions.to_dict())

    print('Start consistency check')

    ltest = line_with_expressions
    lref = line_no_expressions

    ltest.merge_consecutive_drifts()
    lref.merge_consecutive_drifts()

    # Check that the two machines are identical
    assert len(ltest) == len(lref)

    assert (ltest.get_length() - lref.get_length()) < 1e-6

    for ii, (ee_test, ee_six, nn_test, nn_six) in enumerate(
        zip(ltest.elements, lref.elements,
            ltest.element_names, lref.element_names)
    ):
        assert type(ee_test) == type(ee_six)

        dtest = ee_test.to_dict()
        dref = ee_six.to_dict()

        ee_test_cpu = ee_test.copy(_context=xo.ContextCpu())
        ee_six_cpu = ee_six.copy(_context=xo.ContextCpu())

        skip_order = False
        if isinstance(ee_test, xt.Multipole):
            if ee_test._order != ee_six._order:
                min_order = min(ee_test._order, ee_six._order)
                xo.assert_allclose(
                    ee_test.knl[:min_order+1],
                    ee_six.knl[:min_order+1],
                    atol=1e-16,
                )
                xo.assert_allclose(
                    ee_test.ksl[:min_order+1],
                    ee_six.ksl[:min_order+1],
                    atol=1e-16,
                )
                xo.assert_allclose(ee_test.knl[min_order+1:], 0, atol=1e-16)
                xo.assert_allclose(ee_test.ksl[min_order+1:], 0, atol=1e-16)
                xo.assert_allclose(ee_six.knl[min_order+1:], 0, atol=1e-16)
                xo.assert_allclose(ee_six.ksl[min_order+1:], 0, atol=1e-16)

                skip_order = True

        for kk in dtest.keys():

            if skip_order and kk in ('order', 'knl', 'ksl'):
                continue

            # Check if they are identical
            if np.isscalar(dref[kk]) and dtest[kk] == dref[kk]:
                continue

            if isinstance(dref[kk], dict):
                if kk=='fieldmap':
                    continue
                if kk=='boost_parameters':
                    continue
                if kk=='Sigmas_0_star':
                    continue

            # Check if the relative error is small
            val_test = dtest[kk]
            val_ref = dref[kk]
            if not np.isscalar(val_ref):
                if len(val_ref) != len(val_test):
                    diff_rel = 100
                else:
                    for iiii, (vvr, vvt) in enumerate(list(zip(val_ref, val_test))):
                        if not np.isclose(vvr, vvt, atol=atol, rtol=rtol):
                            print(f'Issue found on `{kk}[{iiii}]`')
                            diff_rel = 1000
                        else:
                            diff_rel = 0
            else:
                if val_ref > 0:
                    diff_rel = np.abs((val_test - val_ref)/val_ref)
                else:
                    diff_rel = 100
            if diff_rel < rtol:
                continue

            # Check if absolute error is small
            if not np.isscalar(val_ref) and len(val_ref) != len(val_test):
                diff_abs = 1000
            else:
                diff_abs = norm(np.array(val_test) - np.array(val_ref))
            if diff_abs < atol:
                continue

            # If it got here it means that no condition above is met
            raise ValueError("Too large discrepancy!")

    print('\nTest tracker and xsuite vars...\n')
    line = line_with_expressions.copy()
    line.build_tracker(_context=test_context)
    xo.assert_allclose(line.twiss()['qx'], 62.31, rtol=0, atol=1e-4)
    line.vars['kqtf.b1'] = -2e-4
    xo.assert_allclose(line.twiss()['qx'], 62.2834, rtol=0, atol=1e-4)

    xo.assert_allclose(line.element_dict['acsca.b5l4.b1'].voltage,
                      2e6, rtol=0, atol=1e-14)
    line.vars['vrf400'] = 8
    xo.assert_allclose(line.element_dict['acsca.b5l4.b1'].voltage,
                      1e6, rtol=0, atol=1e-14)

    xo.assert_allclose(line.element_dict['acsca.b5l4.b1'].lag, 180,
                    rtol=0, atol=1e-14)
    line.vars['lagrf400.b1'] = 0.75
    xo.assert_allclose(line.element_dict['acsca.b5l4.b1'].lag, 270,
                    rtol=0, atol=1e-14)

    assert np.abs(
        line.element_dict['acfcav.bl5.b1'].to_dict()['ksl'][0]) > 0
    line.vars['on_crab5'] = 0
    assert np.abs(
        line.element_dict['acfcav.bl5.b1'].to_dict()['ksl'][0]) == 0

    xo.assert_allclose(
        line.element_dict['acfcav.bl5.b1'].to_dict()['ps'][0], 90,
        rtol=0, atol=1e-14)
    line.vars['phi_crab_l5b1'] = 0.5
    xo.assert_allclose(
        line.element_dict['acfcav.bl5.b1'].to_dict()['ps'][0], 270,
                    rtol=0, atol=1e-14)

    assert np.abs(
        line.element_dict['acfcah.bl1.b1'].to_dict()['knl'][0]) > 0
    line.vars['on_crab1'] = 0
    assert np.abs(
        line.element_dict['acfcah.bl1.b1'].to_dict()['knl'][0]) == 0

    xo.assert_allclose(
        line.element_dict['acfcah.bl1.b1'].to_dict()['pn'][0], 90,
        rtol=0, atol=1e-14)
    line.vars['phi_crab_l1b1'] = 0.5
    xo.assert_allclose(
        line.element_dict['acfcah.bl1.b1'].to_dict()['pn'][0], 270,
        rtol=0, atol=1e-14)

    assert np.abs(line.element_dict['acfcah.bl1.b1'].frequency) > 0
    assert np.abs(line.element_dict['acfcav.bl5.b1'].frequency) > 0
    line.vars['crabrf'] = 0.
    assert np.abs(line.element_dict['acfcah.bl1.b1'].frequency) == 0
    assert np.abs(line.element_dict['acfcav.bl5.b1'].frequency) == 0


@for_all_test_contexts
def test_orbit_knobs(test_context, mad_b12_no_errors):

    mad = mad_b12_no_errors

    line = xt.Line.from_madx_sequence(
        mad.sequence['lhcb1'], apply_madx_errors=False,
        deferred_expressions=True)
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1,
                        gamma0=mad.sequence.lhcb1.beam.gamma)

    line.build_tracker(_context=test_context)

    line.vars['on_x1'] = 250
    xo.assert_allclose(line.twiss(at_elements=['ip1'])['px'][0], 250e-6,
                atol=1e-6, rtol=0)
    line.vars['on_x1'] = -300
    xo.assert_allclose(line.twiss(at_elements=['ip1'])['px'][0], -300e-6,
                atol=1e-6, rtol=0)

    line.vars['on_x5'] = 130
    xo.assert_allclose(line.twiss(at_elements=['ip5'])['py'][0], 130e-6,
                atol=1e-6, rtol=0)
    line.vars['on_x5'] = -270
    xo.assert_allclose(line.twiss(at_elements=['ip5'])['py'][0], -270e-6,
                atol=1e-6, rtol=0)


@for_all_test_contexts
def test_low_beta_twiss(test_context):

    path_line = test_data_folder / 'psb_injection/line_and_particle.json'

    line = xt.Line.from_json(path_line)
    line.build_tracker(_context=test_context)
    tw = line.twiss()

    path_madseq = test_data_folder / 'psb_injection/psb_injection.seq'

    mad = Madx(stdout=False)
    mad.call(str(path_madseq))

    mad.use(sequence='psb')
    mad.twiss()
    mad.emit()

    emitdf = mad.table.emitsumm.dframe()

    xo.assert_allclose(mad.sequence.psb.beam.gamma, line.particle_ref.gamma0,
                      rtol=0, atol=1e-6)
    xo.assert_allclose(mad.sequence.psb.beam.beta, line.particle_ref.beta0,
                    rtol=0, atol=1e-10)

    beta0 = line.particle_ref.beta0

    xo.assert_allclose(mad.table.summ['q1'][0], tw['qx'], rtol=0, atol=1e-6)
    xo.assert_allclose(mad.table.summ['q2'][0], tw['qy'], rtol=0, atol=1e-6)
    xo.assert_allclose(mad.table.summ['dq1'][0]*beta0, tw['dqx'], rtol=0,
                        atol=1e-6)
    xo.assert_allclose(mad.table.summ['dq2'][0]*beta0, tw['dqy'], rtol=0,
                        atol=1e-6)
    xo.assert_allclose(tw.qs, emitdf.qs.iloc[0], rtol=0, atol=1e-8)



import xtrack as xt
import numpy as np

magnet = xt.Magnet(k0=0.02, h=0.01, k1=0.01, length=2.,
                   k2=0.005, k3=0.03,
                   k1s=0.01, k2s=0.005, k3s=0.05,
                   knl=[0.003, 0.001, 0.01, 0.02, 4., 6e2, 7e6],
                   ksl=[-0.005, 0.002, -0.02, 0.03, -2, 700., 4e6])

p0 = xt.Particles(x=1e-2, y=2e-2, py=1e-3, delta=3e-2)

m_ref = magnet.copy()
m_ref.model = 'mat-kick-mat'
m_ref.num_multipole_kicks = 1000
p_ref = p0.copy()
m_ref.track(p_ref)

m_uniform = magnet.copy()
# m_uniform.knl[-6] = 0
m_uniform.model = 'drift-kick-drift-expanded'
m_uniform.integrator='uniform'

m_teapot = magnet.copy()
m_teapot.model = 'drift-kick-drift-expanded'
m_teapot.integrator='teapot'

m_yoshida = magnet.copy()
m_yoshida.model = 'drift-kick-drift-expanded'
m_yoshida.integrator='yoshida4'

num_kicks = [1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 2000, 5000, 10000, 20000, 50000,
             100000, 200000, 500000]
log_data = {'num_kicks': [], 'x_teapot': [], 'px_teapot': [], 'y_teapot': [], 'py_teapot': [], 'zeta_teapot': [], 'delta_teapot': [],
            'x_uniform': [], 'px_uniform': [], 'y_uniform': [], 'py_uniform': [], 'zeta_uniform': [], 'delta_uniform': [],
            'x_yoshida': [], 'px_yoshida': [], 'y_yoshida': [], 'py_yoshida': [], 'zeta_yoshida': [], 'delta_yoshida': []}

for nn in num_kicks:
    m_teapot.num_multipole_kicks = nn
    m_uniform.num_multipole_kicks = nn
    m_yoshida.num_multipole_kicks = nn

    p_teapot = p0.copy()
    m_teapot.track(p_teapot)

    p_uniform = p0.copy()
    m_uniform.track(p_uniform)

    p_yoshida = p0.copy()
    m_yoshida.track(p_yoshida)

    # Log data for teapot integrator
    log_data['num_kicks'].append(nn)
    log_data['x_teapot'].append(p_teapot.x[0])
    log_data['px_teapot'].append(p_teapot.px[0])
    log_data['y_teapot'].append(p_teapot.y[0])
    log_data['py_teapot'].append(p_teapot.py[0])
    log_data['zeta_teapot'].append(p_teapot.zeta[0])
    log_data['delta_teapot'].append(p_teapot.delta[0])

    # Log data for uniform integrator
    log_data['x_uniform'].append(p_uniform.x[0])
    log_data['px_uniform'].append(p_uniform.px[0])
    log_data['y_uniform'].append(p_uniform.y[0])
    log_data['py_uniform'].append(p_uniform.py[0])
    log_data['zeta_uniform'].append(p_uniform.zeta[0])
    log_data['delta_uniform'].append(p_uniform.delta[0])

    # Log data for yoshida integrator
    log_data['x_yoshida'].append(p_yoshida.x[0])
    log_data['px_yoshida'].append(p_yoshida.px[0])
    log_data['y_yoshida'].append(p_yoshida.y[0])
    log_data['py_yoshida'].append(p_yoshida.py[0])
    log_data['zeta_yoshida'].append(p_yoshida.zeta[0])
    log_data['delta_yoshida'].append(p_yoshida.delta[0])

for key in log_data.keys():
    log_data[key] = np.array(log_data[key])

import matplotlib.pyplot as plt

plt.close('all')
plt.figure(1, figsize=(12, 8))
ax1 = plt.subplot(2, 3, 1)
plt.loglog(log_data['num_kicks'], np.abs(log_data['x_uniform'] - p_ref.x[0]), '.-', label='uniform')
plt.loglog(log_data['num_kicks'], np.abs(log_data['x_teapot'] - p_ref.x[0]), '.-', label='teapot')
plt.loglog(log_data['num_kicks'], np.abs(log_data['x_yoshida'] - p_ref.x[0]), '.-', label='yoshida')
plt.xlabel('num_kicks')
plt.ylabel('Error in x')
plt.legend()

ax2 = plt.subplot(2, 3, 2, sharex=ax1)
plt.loglog(log_data['num_kicks'], np.abs(log_data['y_uniform'] - p_ref.y[0]), '.-', label='uniform')
plt.loglog(log_data['num_kicks'], np.abs(log_data['y_teapot'] - p_ref.y[0]), '.-', label='teapot')
plt.loglog(log_data['num_kicks'], np.abs(log_data['y_yoshida'] - p_ref.y[0]), '.-', label='yoshida')
plt.xlabel('num_kicks')
plt.ylabel('Error in y')

ax3 = plt.subplot(2, 3, 3, sharex=ax1)
plt.loglog(log_data['num_kicks'], np.abs(log_data['zeta_uniform'] - p_ref.zeta[0]), '.-', label='uniform')
plt.loglog(log_data['num_kicks'], np.abs(log_data['zeta_teapot'] - p_ref.zeta[0]), '.-', label='teapot')
plt.loglog(log_data['num_kicks'], np.abs(log_data['zeta_yoshida'] - p_ref.zeta[0]), '.-', label='yoshida')
plt.xlabel('num_kicks')
plt.ylabel('Error in zeta')

ax4 = plt.subplot(2,3,4, sharex=ax1)
plt.loglog(log_data['num_kicks'], np.abs(log_data['px_uniform'] - p_ref.px[0]), '.-', label='uniform')
plt.loglog(log_data['num_kicks'], np.abs(log_data['px_teapot'] - p_ref.px[0]), '.-', label='teapot')
plt.loglog(log_data['num_kicks'], np.abs(log_data['px_yoshida'] - p_ref.px[0]), '.-', label='yoshida')
plt.xlabel('num_kicks')
plt.ylabel('Error in px')

ax5 = plt.subplot(2,3, 5, sharex=ax1)
plt.loglog(log_data['num_kicks'], np.abs(log_data['py_uniform'] - p_ref.py[0]), '.-', label='uniform')
plt.loglog(log_data['num_kicks'], np.abs(log_data['py_teapot'] - p_ref.py[0]), '.-', label='teapot')
plt.loglog(log_data['num_kicks'], np.abs(log_data['py_yoshida'] - p_ref.py[0]), '.-', label='yoshida')
plt.xlabel('num_kicks')
plt.ylabel('Error in py')

ax6 = plt.subplot(2, 3, 6, sharex=ax1)
# plt.loglog(log_data['num_kicks'], np.abs(log_data['delta_uniform'] - p_ref.delta[0]), '.-', label='uniform')
# plt.loglog(log_data['num_kicks'], np.abs(log_data['delta_teapot'] - p_ref.delta[0]), '.-', label='teapot')
# plt.loglog(log_data['num_kicks'], np.abs(log_data['delta_yoshida'] - p_ref.delta[0]), '.-', label='yoshida')
# plt.xlabel('num_kicks')

plt.subplots_adjust(hspace=0.3, wspace=0.3)
plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time
import numpy as np

from cpymad.madx import Madx

import xtrack as xt

# Import a thick sequence
mad = Madx()
mad.call('../../test_data/clic_dr/sequence.madx')
mad.use('ring')

# Makethin
mad.input(f'''
select, flag=MAKETHIN, SLICE=4, thick=false;
select, flag=MAKETHIN, pattern=wig, slice=1;
select, flag=makethin, class=rfcavity, slice=1;
MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
use, sequence=RING;
''')
mad.use('ring')

# Build xtrack line
print('Build xtrack line...')
line = xt.Line.from_madx_sequence(mad.sequence['RING'])
line.particle_ref = xt.Particles(
        mass0=xt.ELECTRON_MASS_EV,
        q0=-1,
        gamma0=mad.sequence.ring.beam.gamma)

# Build tracker
print('Build tracker ...')
line.build_tracker()

################################
# Enable synchrotron radiation #
################################

# we choose the `mean` mode in which the mean power loss is applied without
# stochastic fluctuations (quantum excitation).
line.configure_radiation(model='mean')

#########
# Twiss #
#########

tw = line.twiss(eneloss_and_damping=True)

# By setting `eneloss_and_damping=True` we can get additional information
# from the twiss for example:
#  - tw['eneloss_turn'] provides the energy loss per turn (in eV).
#  - tw['damping_constants_s'] provides the damping constants in x, y and zeta.
#  - tw['partition_numbers'] provides the corresponding damping partion numbers.
#  - tw['eq_nemitt_x'] provides the equilibrium horizontal emittance.
#  - tw['eq_nemitt_y'] provides the equilibrium vertical emittance.
#  - tw['eq_nemitt_zeta'] provides the equilibrium longitudinal emittance.

############################################
# Generate particles and track (mean mode) #
############################################

# Build three particles (with action in x,y and zeta respectively)
part_co = tw['particle_on_co']
particles = line.build_particles(
    x_norm=[500., 0, 0], y_norm=[0, 500, 0], zeta=part_co.zeta[0],
    delta=np.array([0,0,1e-2]) + part_co.delta[0],
    nemitt_x=1e-9, nemitt_y=1e-9)

# Save initial state
particles_0 = particles.copy()

# Track
num_turns = 5000
line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)

# Save monitor
mon_mean_mode = line.record_last_track

############################
# Switch to `quantum` mode #
############################

# We switch to the `quantum` mode in which the power loss from radiation is
# applied including stochastic fluctuations (quantum excitation).
# IMPORTANT: Note that this mode should not be used to compute twiss parameters
#            nor to match particle distributions. For this reason we switch
#            to quantum mode only after having generated the particles.


line.configure_radiation(model='quantum')

# We reuse the initial state saved before
particles = particles_0.copy()

num_turns = 5000
line.track(particles, num_turns=num_turns, turn_by_turn_monitor=True)
mon_quantum_mode = line.record_last_track

import matplotlib.pyplot as plt
plt.close('all')
figs = []
for ii, mon in enumerate([mon_mean_mode, mon_quantum_mode]):
    fig = plt.figure(ii + 1)
    ax1 = fig.add_subplot(311)
    ax2 = fig.add_subplot(312, sharex=ax1)
    ax3 = fig.add_subplot(313, sharex=ax1)

    ax1.plot(1e3*mon.x[0, :].T)
    ax2.plot(1e3*mon.y[1, :].T)
    ax3.plot(1e3*mon.delta[2, :].T)
    i_turn = np.arange(num_turns)
    ax1.plot(1e3*(part_co.x[0]
        +(mon.x[0,0]-part_co.x[0])*np.exp(-i_turn*tw['damping_constants_turns'][0])))
    ax2.plot(1e3*(part_co.y[0]
        +(mon.y[1,0]-part_co.y[0])*np.exp(-i_turn*tw['damping_constants_turns'][1])))
    ax3.plot(1e3*(part_co.delta[0]
        +(mon.delta[2,0]-part_co.delta[0])*np.exp(-i_turn*tw['damping_constants_turns'][2])))

    ax1.set_ylabel('x [mm]')
    ax2.set_ylabel('y [mm]')
    ax3.set_ylabel('delta [-]')
    ax3.set_xlabel('Turn')

plt.show()




import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

line = env.new_line(
    components=[
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=-10.0, from_='q0'),
        env.new('qr', 'Quadrupole', length=2.0, at=10.0, from_='q0'),
        env.new('mk1', 'Marker', at=40),
        env.new('mk2', 'Marker', at=42),
        env.new('end', 'Marker', at=50.),
    ])

s_tol = 1e-10

tt0 = line.get_table()
tt0.show(cols=['name', 's_start', 's_end', 's_center'])

env.new('q1', 'q0')
env.new('q2', 'q0')
env.new('q3', 'q0')

line.insert('q1', at=5.0)
line.insert('q2', at=15.0)
line.insert('q3', at=41.0)

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1..0', 'q1', 'drift_1..2', 'ql', 'drift_2..0', 'q2',
       'drift_2..2', 'q0', 'drift_3', 'qr', 'drift_4', 'mk1', 'q3', 'mk2',
       'drift_6', 'end', '_end_point']))
xo.assert_allclose(tt.s_center, np.array(
    [ 2. ,  5. ,  7.5, 10. , 12.5, 15. , 17.5, 20. , 25. , 30. , 35.5,
       40. , 41. , 42. , 46. , 50. , 50. ]), rtol=0., atol=1e-14)


import xtrack as xt
import numpy as np

from scipy.constants import hbar
from scipy.constants import electron_volt
from scipy.constants import c as clight

line = xt.Line.from_json('line_thick_P4_H6BA_v4.2.4.json')
line.particle_ref = xt.Particles(energy0=6e9, mass0=xt.ELECTRON_MASS_EV)

tw4d = line.twiss4d()
tw6d = line.twiss()

tt = line.get_table(attr=True)
tt_bend = tt.rows[tt.element_type == 'Bend']
tt_quad = tt.rows[tt.element_type == 'Quadrupole']
tt_sext = tt.rows[tt.element_type == 'Sextupole']

line.set(tt_sext, integrator='yoshida4', num_multipole_kicks=7)
line.set(tt_bend, model='mat-kick-mat', integrator='teapot', num_multipole_kicks=1)
line.set(tt_bend.rows['wgl.*'], model='drift-kick-drift-expanded',
         integrator='teapot', num_multipole_kicks=1)

tw_after = line.twiss4d()

line.configure_radiation(model='mean')
tw_rad = line.twiss(eneloss_and_damping=True, strengths=True)

p = line.build_particles(x=np.linspace(-1e-3, 1e-3, 1000))
line.track(p, num_turns=5, with_progress=1, time=True)
print('Time for tracking: ', line.time_last_track)


import numpy as np
from cpymad.madx import Madx
import xtrack as xt

# We get the model from MAD-X
mad = Madx()
folder = ('../../test_data/elena')
mad.call(folder + '/elena.seq')
mad.call(folder + '/highenergy.str')
mad.call(folder + '/highenergy.beam')
mad.use('elena')

# Build xsuite line
seq = mad.sequence.elena
line = xt.Line.from_madx_sequence(seq)
line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
                                    mass0=seq.beam.mass * 1e9,
                                    q0=seq.beam.charge)

# Inspect one bend
line['lnr.mbhek.0135']
# returns:
#
# Bend(length=0.971, k0=1.08, k1=0, h=1.08, model='adaptive',
#      knl=array([0., 0., 0., 0., 0.]), ksl=array([0., 0., 0., 0., 0.]),
#      edge_entry_active=1, edge_exit_active=1,
#      edge_entry_model='linear', edge_exit_model='linear',
#      edge_entry_angle=0.287, edge_exit_angle=0.287,
#      edge_entry_angle_fdown=0, edge_exit_angle_fdown=0,
#      edge_entry_fint=0.424, edge_exit_fint=0.424,
#      edge_entry_hgap=0.038, edge_exit_hgap=0.038,
#      shift_x=0, shift_y=0, rot_s_rad=0)

# By default the adaptive model is used for the core and the linearized model for the edge
line['lnr.mbhek.0135'].model # is 'adaptive'
line['lnr.mbhek.0135'].edge_entry_model # is 'linear'
line['lnr.mbhek.0135'].edge_exit_model # is 'linear'

# For small machines (bends with large bending angles) it is more appropriate to
# switch to the `full` model for the edge
line.configure_bend_model(core='adaptive', edge='full')

# It is also possible to switch from the expanded drift to the exact one
line.config.XTRACK_USE_EXACT_DRIFTS = True

line['lnr.mbhek.0135'].model # is 'adaptive'
line['lnr.mbhek.0135'].edge_entry_model # is 'full'
line['lnr.mbhek.0135'].edge_exit_model # is 'full'

# Slice the bends to see the behavior of the optics functions within them
line.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(slicing=None), # don't touch other elements
        xt.Strategy(slicing=xt.Uniform(10, mode='thick'), element_type=xt.Bend)
    ])

# Twiss
tw = line.twiss(method='4d')

# Switch to a simplified model
line.configure_bend_model(core='expanded', edge='linear')
line.config.XTRACK_USE_EXACT_DRIFTS = False

# Twiss with the default model
tw_simpl = line.twiss(method='4d')

# Compare beta functions and chromatic properties

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1, figsize=(6.4, 4.8 * 1.5))
ax1 = plt.subplot(4,1,1)
plt.plot(tw.s, tw.betx, label='adaptive')
plt.plot(tw_simpl.s, tw_simpl.betx, '--', label='simplified')
plt.ylabel(r'$\beta_x$')
plt.legend(loc='best')

ax2 = plt.subplot(4,1,2, sharex=ax1)
plt.plot(tw.s, tw.bety)
plt.plot(tw_simpl.s, tw_simpl.bety, '--')
plt.ylabel(r'$\beta_y$')

ax3 = plt.subplot(4,1,3, sharex=ax1)
plt.plot(tw.s, tw.wx_chrom)
plt.plot(tw_simpl.s, tw_simpl.wx_chrom, '--')
plt.ylabel(r'$W_x$')

ax4 = plt.subplot(4,1,4, sharex=ax1)
plt.plot(tw.s, tw.wy_chrom)
plt.plot(tw_simpl.s, tw_simpl.wy_chrom, '--')
plt.ylabel(r'$W_y$')
plt.xlabel('s [m]')

# Highlight the bends
tt_sliced = line.get_table()
tbends = tt_sliced.rows[tt_sliced.element_type == 'ThickSliceBend']
for ax in [ax1, ax2, ax3, ax4]:
    for nn in tbends.name:
        ax.axvspan(tbends['s', nn], tbends['s', nn] + line[nn].length,
                   color='b', alpha=0.2, linewidth=0)

plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from .general import _pkg_root, _print, START, END

from .particles import (Particles, PROTON_MASS_EV, ELECTRON_MASS_EV,
                        enable_pyheadtail_interface, disable_pyheadtail_interface)

from .base_element import BeamElement, Replica
from .beam_elements import *
from .random import *
from .tracker_data import TrackerData
from .line import Line, Node, freeze_longitudinal, _temp_knobs, EnergyProgram
from .environment import Environment, Place, get_environment
from .tracker import Tracker, Log
from .match import (Vary, Target, TargetList, VaryList, TargetInequality, Action,
                    TargetRelPhaseAdvance, TargetSet, GreaterThan, LessThan,
                    TargetRmatrixTerm, TargetRmatrix)
from .targets import (TargetLuminosity, TargetSeparationOrthogonalToCrossing,
                      TargetSeparation)
from .twiss import TwissInit, TwissTable
from .loss_location_refinement import LossLocationRefinement
from .internal_record import (RecordIdentifier, RecordIndex, new_io_buffer,
                             start_internal_logging, stop_internal_logging)
from .pipeline import (PipelineStatus, PipelineMultiTracker, PipelineBranch,
                        PipelineManager)

from .monitors import *
from . import linear_normal_form
from .multiline_legacy import MultilineLegacy, MultiTwiss

from .mad_loader import MadLoader

from .multisetter import MultiSetter

from .footprint import Footprint, LinearRescale

# Flags and test functions
from .line import _is_drift, _behaves_like_drift, _is_aperture, _is_thick, _allow_loss_refinement
from .line import _lines_equal, _apertures_equal

from .slicing import Strategy, Uniform, Teapot
from .loss_location_refinement import _skip_in_loss_location_refinement
from .trajectory_correction import TrajectoryCorrection
from .mad_parser.loader import load_madx_lattice
from . import json

from .multiline import Multiline

from xdeps import Table, FunctionPieceWiseLinear

from ._version import __version__





def test_hvkick():
    import xtrack as xt
    from cpymad.madx import Madx

    # Load a very simple sequence from MAD-X
    mad = Madx()
    mad.input("""
        seq: sequence, l=4;
        b1: sbend, at=0.5, angle=0.2, l=1;
        b2: sbend, at=2.5, angle=0.3, k0=0.15, l=1;
        m1: multipole, at=3, knl={0.1};
        k1: kicker, at=3, hkick=0.3, vkick=0.4;
        endsequence;
        beam;
        use,sequence=seq;
    """)

    line = xt.Line.from_madx_sequence(mad.sequence.seq)
    line.build_tracker()

    print('The line as imported from MAD-X:')
    line.get_table().show()

    # Shift and tilt selected elements
    line['b1'].rot_s_rad = 0.8
    line['b2'].rot_s_rad = -0.8

    tt = line.get_table(attr=True)
    tt.cols['s', 'element_type', 'angle_rad', 'rot_s_rad', 'k0l', 'k0sl', 'hkick', 'vkick']

    assert  tt['hkick','k1']==0.3
    assert  tt['vkick','k1']==0.4
    assert  tt['hkick','m1']==0
    assert  tt['vkick','m1']==0
    assert  tt['hkick','b1']==0
    assert  tt['vkick','b1']==0
    assert  tt['hkick','b2']==0.15
    assert  tt['vkick','b2']==0




import numpy as np

import xobjects as xo
import xtrack as xt

## Generate a simple line
line = xt.Line(
    elements=[xt.Drift(length=2.),
              xt.Multipole(knl=[0, 0.5], ksl=[0,0]),
              xt.Drift(length=1.),
              xt.Multipole(knl=[0, -0.5], ksl=[0,0])],
    element_names=['drift_0', 'quad_0', 'drift_1', 'quad_1'])

## Attach a reference particle to the line (optional)
## (defines the reference mass, charge and energy)
line.particle_ref = xt.Particles(p0c=6500e9, #eV
                                 q0=1, mass0=xt.PROTON_MASS_EV)

## Choose a context
context = xo.ContextCpu()         # For CPU
# context = xo.ContextCupy()      # For CUDA GPUs
# context = xo.ContextPyopencl()  # For OpenCL GPUs

## Transfer lattice on context and compile tracking code
line.build_tracker(_context=context)

## Build particle object on context
n_part = 20
particles = line.build_particles(
                        x=np.random.uniform(-1e-3, 1e-3, n_part),
                        px=np.random.uniform(-1e-5, 1e-5, n_part),
                        y=np.random.uniform(-2e-3, 2e-3, n_part),
                        py=np.random.uniform(-3e-5, 3e-5, n_part),
                        zeta=np.random.uniform(-1e-2, 1e-2, n_part),
                        delta=np.random.uniform(-1e-4, 1e-4, n_part))
# Reference mass, charge, energy are taken from the reference particle.
# Particles are allocated on the context chosen for the line.

## Track (no saving of turn-by-turn data)
n_turns = 100
line.track(particles, num_turns=n_turns)

particles.state # > 0 for particles still alive
particles.at_turn # turn number (for lost particles, it is the turn of loss)
particles.x # x position after tracking
particles.px # x momentum after tracking
# etc...

## Track (saving turn-by-turn data)
particles = line.build_particles( # fresh particles
                        x=np.random.uniform(-1e-3, 1e-3, n_part),
                        px=np.random.uniform(-1e-5, 1e-5, n_part),
                        y=np.random.uniform(-2e-3, 2e-3, n_part),
                        py=np.random.uniform(-3e-5, 3e-5, n_part),
                        zeta=np.random.uniform(-1e-2, 1e-2, n_part),
                        delta=np.random.uniform(-1e-4, 1e-4, n_part))
n_turns = 100
line.track(particles, num_turns=n_turns,
              turn_by_turn_monitor=True)

## Turn-by-turn data is available at:
line.record_last_track.x
line.record_last_track.px
# etc...


import pathlib

import numpy as np
from scipy.constants import c as clight

import xtrack as xt
from xobjects.test_helpers import fix_random_seed

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@fix_random_seed(8837465)
def test_coasting():
    delta0 = 1e-2

    line = xt.Line.from_json(test_data_folder /
                             'psb_injection/line_and_particle.json')

    # RF off!
    tt = line.get_table()
    ttcav = tt.rows[tt.element_type == 'Cavity']
    for nn in ttcav.name:
        line.element_refs[nn].voltage=0

    line.configure_bend_model(core='bend-kick-bend', edge='full')
    line.twiss_default['method'] = '4d'

    tw = line.twiss()
    twom = line.twiss(delta0=delta0)
    line.discard_tracker()

    # Install dummy collective elements
    s_sync = np.linspace(0, tw.circumference, 10)
    line.cut_at_s(s_sync)
    for ii, ss in enumerate(s_sync):
        nn = f'sync_here_{ii}'
        line.insert_element(element=xt.Marker(), name=nn, at_s=ss)
        line[nn].iscollective = True

    import xtrack.synctime as st
    st.install_sync_time_at_collective_elements(line)

    import xobjects as xo
    line.build_tracker(_context=xo.ContextCpu(omp_num_threads='auto'))

    beta1 = tw.beta0 / 0.9

    circumference = tw.circumference
    zeta_min0 = -circumference/2*tw.beta0/beta1
    zeta_max0 = circumference/2*tw.beta0/beta1

    num_particles = 50000
    p = line.build_particles(
        delta=delta0 + 0 * np.random.uniform(-1, 1, num_particles),
        x_norm=0, y_norm=0
    )

    # Need to take beta of actual particles to convert the distribution along the
    # circumference to a distribution in time
    p.zeta = (np.random.uniform(0, circumference, num_particles) / p.rvv
            + (zeta_max0 - circumference) / p.rvv)

    st.prepare_particles_for_sync_time(p, line)

    p.y[(p.zeta > 1) & (p.zeta < 2)] = 1e-3  # kick
    p.weight[(p.zeta > 5) & (p.zeta < 10)] *= 1.3

    p0 = p.copy()

    def particles(_, p):
        return p.copy()

    def intensity(line, particles):
        return (np.sum(particles.weight[particles.state > 0])
                    / ((zeta_max0 - zeta_min0)/tw.beta0/clight))

    def z_range(line, particles):
        mask_alive = particles.state > 0
        return particles.zeta[mask_alive].min(), particles.zeta[mask_alive].max()

    def long_density(line, particles):
        mask_alive = particles.state > 0
        if not(np.any(particles.at_turn[mask_alive] == 0)): # don't check at the first turn
            assert np.all(particles.zeta[mask_alive] > zeta_min0)
            assert np.all(particles.zeta[mask_alive] < zeta_max0)
        return np.histogram(particles.zeta[mask_alive], bins=200,
                            range=(zeta_min0, zeta_max0),
                            weights=particles.weight[mask_alive])

    def y_mean_hist(line, particles):

        mask_alive = particles.state > 0
        if not(np.any(particles.at_turn[mask_alive] == 0)): # don't check at the first turn
            assert np.all(particles.zeta[mask_alive] > zeta_min0)
            assert np.all(particles.zeta[mask_alive] < zeta_max0)
        return np.histogram(particles.zeta[mask_alive], bins=200,
                            range=(zeta_min0, zeta_max0), weights=particles.y[mask_alive])


    line.enable_time_dependent_vars = True
    num_turns=200
    line.track(p, num_turns=num_turns, log=xt.Log(intensity=intensity,
                                            long_density=long_density,
                                            y_mean_hist=y_mean_hist,
                                            z_range=z_range,
                                            particles=particles
                                            ), with_progress=10)

    inten = line.log_last_track['intensity']

    f_rev_ave = 1 / tw.T_rev0 * (1 - tw.slip_factor * p.delta.mean())
    t_rev_ave = 1 / f_rev_ave

    inten_exp =  np.sum(p0.weight) / t_rev_ave

    z_axis = line.log_last_track['long_density'][0][1]
    hist_mat = np.array([rr[0] for rr in line.log_last_track['long_density']])
    hist_y = np.array([rr[0] for rr in line.log_last_track['y_mean_hist']])

    dz = z_axis[1] - z_axis[0]
    y_vs_t = np.fliplr(hist_y).flatten() # need to flip because of the minus in z = -beta0 c t
    intensity_vs_t = np.fliplr(hist_mat).flatten()
    z_unwrapped = np.arange(0, len(y_vs_t)) * dz
    t_unwrapped = z_unwrapped / (tw.beta0 * clight)

    z_range_size = z_axis[-1] - z_axis[0]
    t_range_size = z_range_size / (tw.beta0 * clight)

    import nafflib
    intensity_no_ave = intensity_vs_t - np.mean(intensity_vs_t)
    f_harmons = nafflib.get_tunes(intensity_no_ave, N=50)[0] / (t_unwrapped[1] - t_unwrapped[0])
    f_nominal = 1 / tw.T_rev0
    dt_expected = -(twom.zeta[-1] - twom.zeta[0]) / tw.beta0 / clight
    f_expected = 1 / (tw.T_rev0 + dt_expected)

    f_measured = f_harmons[np.argmin(np.abs(f_harmons - f_nominal))]

    print('f_nominal:  ', f_nominal, ' Hz')
    print('f_expected: ', f_expected, ' Hz')
    print('f_measured: ', f_measured, ' Hz')
    print('Error:      ', f_measured - f_expected, 'Hz')

    xo.assert_allclose(f_expected, f_measured, rtol=0, atol=5.) # 5 Hz tolerance (to account for random fluctuations)
    xo.assert_allclose(np.mean(inten), inten_exp, rtol=1e-2, atol=0)
    xo.assert_allclose(p.at_turn, num_turns*0.9, rtol=3e-2, atol=0) #beta1 defaults to 0.1

    tt = line.get_table()
    tt_synch = tt.rows[tt.element_type=='SyncTime']
    assert len(tt_synch) == 12
    assert tt_synch.name[0] == 'synctime_start'
    assert tt_synch.name[-1] == 'synctime_end'
    assert np.all(tt_synch.name[5] == 'synctime_4')
    assert line['synctime_start'].at_start
    assert not line['synctime_end'].at_start
    assert not line['synctime_4'].at_start
    assert line['synctime_end'].at_end
    assert not line['synctime_start'].at_end
    assert not line['synctime_4'].at_end



import numpy as np

from cpymad.madx import Madx
import xtrack as xt
import xdeps as xd

# Load a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, p0c=7e12)
line.build_tracker()

# Cycle the line to the middle of the bump
line.cycle('ip1', inplace=True)
line.build_tracker()

line.vars['on_x1'] = 0
tw = line.twiss(method='4d')

# Define the function
line.functions['f_on_sep1'] = xd.FunctionPieceWiseLinear(
    x=np.array([0, 5, 20, 40, 50]) * 1e-3, # ms
    y=np.array([0, 0,  2,  2,  1]) # knob value
)

# Drive the knob with the function
line.vars['on_sep1'] = line.functions.f_on_sep1(line.vars['t_turn_s'])

# Track a particle on the closed orbit
p = line.build_particles(x=0)
line.enable_time_dependent_vars = True
line.track(p, num_turns=1000, turn_by_turn_monitor=True)
mon = line.record_last_track

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(mon.at_turn.T * tw.T_rev0*1e3, mon.y.T)
plt.xlabel('time [ms]')
plt.ylabel('y [m]')
plt.show()






# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import abc
import re
from abc import ABC

from itertools import zip_longest
from typing import List, Tuple, Iterator, Optional, Literal

import numpy as np

from .progress_indicator import progress

import xtrack as xt

APER_ELEMS_REGEX = re.compile(r'(.*)_aper(_(tilt|offset)_(entry|exit))?')


class ElementSlicingScheme(abc.ABC):
    def __init__(
            self,
            slicing_order: int,
            mode: Literal['thin', 'thick'] = 'thin',
    ):
        if slicing_order < 1:
            raise ValueError("A slicing scheme must have at least one slice.")

        self.mode = mode

        if mode == 'thick':
            # In thick mode we count the number of thick slices, 'drifts',
            # not thin ones, so we need to decrement the slicing order.
            slicing_order -= 1

        self.slicing_order = slicing_order

    @abc.abstractmethod
    def element_weights(self, element_length: float) -> List[float]:
        """Define a list of weights of length `self.slicing_order`, containing
         the weight of each element slice.
        """

    @abc.abstractmethod
    def drift_weights(self, element_length: float) -> List[float]:
        """Define a list of weights of length `self.slicing_order + 1`,
        containing the weight of each drift slice.
        """

    def iter_weights(
            self,
            element_length: float = None,
    ) -> Iterator[Tuple[float, bool]]:
        """
        Give an iterator for weights of slices and, assuming the first slice is
        a drift, followed by an element slice, and so on.
        Returns
        -------
        Iterator[Tuple[float, bool]]
            Iterator of weights and whether the weight is for a drift.
        """
        for drift_weight, elem_weight in zip_longest(
                self.drift_weights(element_length),
                self.element_weights(element_length),
                fillvalue=None,
        ):
            yield drift_weight, True

            if elem_weight is None:
                break

            if self.mode == 'thin':
                yield elem_weight, False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.slicing_order})"


class Uniform(ElementSlicingScheme):
    def element_weights(self, element_length=None):
        if self.slicing_order == 0 and self.mode == 'thick':
            return [1.]

        return [1. / self.slicing_order] * self.slicing_order

    def drift_weights(self, element_length=None):
        if self.slicing_order == 0:
            return [1.]
        slices = self.slicing_order + 1
        return [1. / slices] * slices


class Teapot(ElementSlicingScheme):
    def element_weights(self, element_length=None):
        if self.slicing_order == 0 and self.mode == 'thick':
            return [1.]

        return [1. / self.slicing_order] * self.slicing_order

    def drift_weights(self, element_length=None):

        if self.slicing_order == 0:
            return [1.]

        if self.slicing_order == 1:
            return [0.5, 0.5]

        edge_weight = 1. / (2 * (1 + self.slicing_order))
        middle_weight = self.slicing_order / (self.slicing_order ** 2 - 1)
        middle_weights = [middle_weight] * (self.slicing_order - 1)

        return [edge_weight, *middle_weights, edge_weight]


class Custom(ElementSlicingScheme):
    """The custom slicing scheme slices the element at the fixed s coordinates.

    Arguments
    ---------
    at_s
        The s values at which the elements should be sliced. The beginning of
        the element is assumed to be at zero.
    mode:
        Thick or thin slicing.
    """
    def __init__(
            self,
            at_s: List[float],
            mode: Literal['thin', 'thick'] = 'thick',
    ):
        slicing_order = len(at_s)
        if mode == 'thick':
            # In thick number of slices is one more than cuts
            slicing_order = len(at_s) + 1

        super().__init__(slicing_order, mode)

        # Precompute the known weights (all but the last slice, which depends
        # on the weight of the element being sliced)
        self.at_s = np.array([0.] + at_s)
        self.section_lengths = self.at_s[1:] - self.at_s[:-1]

    def element_weights(self, element_length=None):
        return [1. / self.slicing_order] * self.slicing_order

    def drift_weights(self, element_length: float) -> List[float]:
        section_lengths = np.concatenate([
            self.section_lengths,
            [element_length - self.at_s[-1]],
        ])
        weights = section_lengths / element_length
        return weights.tolist()

    def __repr__(self):
        return f'Custom(at_s={self.at_s.tolist()})'


class Strategy:
    def __init__(self, slicing, name=None, element_type=None, exact=False):
        if name is None and element_type is None:
            exact = True

        self.regex = not exact

        if name is not None and isinstance(name, str) and self.regex:
            self.match_name = re.compile(name)
        else:
            self.match_name = name

        self.element_type = element_type
        self.slicing = slicing

    def _match_on_name(self, name):
        if self.regex:
            return self.match_name.match(name)
        return self.match_name == name

    def _match_on_type(self, element, line):
        if isinstance(element, xt.Replica):
            element = element.resolve(line)
        return isinstance(element, self.element_type)

    def match_element(self, name, element, line):

        if isinstance(element, xt.Drift):
            matched = False
            if self.match_name and not self.element_type:
                matched = self._match_on_name(name)
            elif self.element_type and not self.match_name:
                matched = self._match_on_type(element, line)
            elif self.match_name and self.element_type:
                matched = self._match_on_name(name) and self._match_on_type(element, line)
        else:
            matched = True
            if self.match_name:
                matched = matched and self._match_on_name(name)
            if self.element_type:
                matched = matched and self._match_on_type(element, line)
        return matched

    def __repr__(self):
        params = {
            'slicing': self.slicing,
            'element_type': self.element_type,
            'name': self.match_name.pattern if self.regex else self.match_name,
        }
        formatted_params = ', '.join(
            f'{kk}={vv!r}' for kk, vv in params.items() if vv is not None
        )
        return f"{type(self).__name__}({formatted_params})"


class Slicer:
    def __init__(self, line, slicing_strategies: List[Strategy]):
        """
        An object that slices a line in place according to a list of slicing
        strategies.

        Parameters
        ----------
        line : Line
            The line to slice.
        slicing_strategies : List[Strategy]
            A list of slicing strategies to apply to the line.
        """

        self._line = line
        self._slicing_strategies = [xt.Strategy(None, element_type=xt.Drift) # Do nothing to drifts by default
                                    ] + slicing_strategies
        self._has_expressions = line.vars is not None

    def slice_in_place(self, _edge_markers=True):

        self._line._frozen_check()

        thin_names = []

        tt = self._line.get_table()
        assert tt.name[-1] == '_end_point'
        tt = tt.rows[:-1]
        slices = {}
        for ii, nn in enumerate(progress(tt.name, desc='Slicing line')):

            enn=tt.env_name[ii]
            element = self._line.element_dict[enn]

            subsequence = [enn]
            # Don't slice already thin elements and drifts
            if (not xt.line._is_thick(element, self._line)
                or (hasattr(element, 'length') and element.length == 0)):
                pass
            else:
                chosen_slicing = self._scheme_for_element(element, nn, self._line)
                if chosen_slicing is not None:
                    subsequence = self._slice_element(
                        enn, element, _edge_markers=_edge_markers,
                        chosen_slicing=chosen_slicing)
                if subsequence is None:
                    subsequence = [enn]

            thin_names += subsequence
            slices[nn] = subsequence

        # Commit the changes to the line
        self._line.element_names = thin_names

        return slices

    def _slice_element(self, name, element, chosen_slicing, _edge_markers=True) -> Optional[List[str]]:
        """Slice element and return slice names, or None if no slicing."""

        if isinstance(element, xt.Drift) or type(element).__name__.startswith('DriftSlice'):
            _edge_markers = False

        # Make the slices and add them to line.element_dict (so far inactive)
        slices_to_add = self._make_slices(
            element=element,
            chosen_slicing=chosen_slicing,
            name=name,
        )

        if _edge_markers:

            if isinstance(element, xt.Replica):
                element = element.resolve(self._line)
            _buffer = element._buffer

            entry_marker, exit_marker = f'{name}_entry', f'{name}_exit'
            self._line.element_dict[entry_marker] = xt.Marker(_buffer=_buffer)
            self._line.element_dict[exit_marker] = xt.Marker(_buffer=_buffer)
            slices_to_add = [entry_marker] + slices_to_add + [exit_marker]

        # Handle aperture
        if isinstance(element, xt.Replica):
            element = element.resolve(self._line)
        if (hasattr(element, 'name_associated_aperture')
            and element.name_associated_aperture is not None):
            new_slices_to_add = []
            aper_index = 0
            for nn in slices_to_add:
                ee = self._line.element_dict[nn]
                if (type(ee).__name__.startswith('ThinSlice')
                    or type(ee).__name__.startswith('ThickSlice')):
                    aper_name = f'{name}_aper..{aper_index}'
                    self._line.element_dict[aper_name] = xt.Replica(
                        parent_name=element.name_associated_aperture)
                    new_slices_to_add += [aper_name]
                    aper_index += 1
                new_slices_to_add += [nn]
            slices_to_add = new_slices_to_add

        return slices_to_add

    def _scheme_for_element(self, element, name, line):
        """Choose a slicing strategy for the element"""

        slicing_found = False
        chosen_slicing = None

        for strategy in reversed(self._slicing_strategies):
            if strategy.match_element(name, element, line):
                slicing_found = True
                chosen_slicing = strategy.slicing
                break
        if not slicing_found:
            raise ValueError(f'No slicing strategy found for the element '
                             f'{name}: {element}.')
        return chosen_slicing

    def _make_slices(self, element, chosen_slicing, name):
        """
        Add the slices to the line.element_dict. If the element has expressions
        then the expressions will be added to the slices.

        Parameters
        ----------
        element : BeamElement
            A thick element to slice.
        chosen_slicing : ElementSlicingScheme
            The slicing scheme to use for the element.
        name : str
            The name of the element.

        Returns
        -------
        list
            A list of the names of the slices that were added.
        """

        parent_name = name
        if isinstance(element, xt.Replica):
            parent_name = element.resolve(self._line, get_name=True)
            element = self._line[parent_name]

        drift_idx, element_idx = 0, 0
        slices_to_append = []

        if hasattr(element, '_entry_slice_class'):
            nn = f'{name}..entry_map'
            if nn in self._line.element_dict:
                i_entry = 0
                while (nn := f'{name}..entry_map_{i_entry}') in self._line.element_dict:
                    i_entry += 1
            ee = element._entry_slice_class(
                    _parent=element, _buffer=element._buffer)
            ee.parent_name = parent_name
            self._line.element_dict[nn] = ee
            slices_to_append.append(nn)

        if not hasattr(element, 'length'):
            # Slicing a thick slice of a another element
            assert hasattr(element, '_parent')
            assert element.isthick
            elem_length = element._parent.length * element.weight
            elem_weight = element.weight
            slice_parent_name = element.parent_name
            slice_parent = element._parent
            isdriftslice = type(element).__name__.startswith('DriftSlice')
        else:
            elem_length = element.length
            elem_weight = 1.
            slice_parent_name = parent_name
            slice_parent = element
            isdriftslice = False

        if chosen_slicing.mode == 'thin' or isdriftslice:
            for weight, is_drift in chosen_slicing.iter_weights(elem_length):
                prename = "" if isdriftslice else "drift_"
                if is_drift:
                    while (nn := f'{prename}{name}..{drift_idx}') in self._line.element_dict:
                        drift_idx += 1
                    ee = slice_parent._drift_slice_class(
                            _parent=slice_parent, _buffer=element._buffer,
                            weight=weight * elem_weight)
                    ee.parent_name = slice_parent_name
                    self._line.element_dict[nn] = ee
                    slices_to_append.append(nn)
                else:
                    if isdriftslice:
                        continue
                    while (nn := f'{name}..{element_idx}') in self._line.element_dict:
                        element_idx += 1
                    if slice_parent._thin_slice_class is not None:
                        ee = slice_parent._thin_slice_class(
                                _parent=slice_parent, _buffer=element._buffer,
                                weight=weight * elem_weight)
                        ee.parent_name = slice_parent_name
                        self._line.element_dict[nn] = ee
                        slices_to_append.append(nn)
        elif chosen_slicing.mode == 'thick':
            for weight, is_drift in chosen_slicing.iter_weights(elem_length):
                while (nn := f'{name}..{element_idx}') in self._line.element_dict:
                    element_idx += 1
                ee = slice_parent._thick_slice_class(
                        _parent=slice_parent, _buffer=element._buffer,
                        weight=weight * elem_weight)
                ee.parent_name = slice_parent_name
                self._line.element_dict[nn] = ee
                slices_to_append.append(nn)
        else:
            raise ValueError(f'Unknown slicing mode: {chosen_slicing.mode}')

        if hasattr(element, '_exit_slice_class'):
            nn = f'{name}..exit_map'
            if nn in self._line.element_dict:
                i_exit = 0
                while (nn := f'{name}..exit_map_{i_exit}') in self._line.element_dict:
                    i_exit += 1
            ee = element._exit_slice_class(
                    _parent=element, _buffer=element._buffer)
            ee.parent_name = parent_name
            self._line.element_dict[nn] = ee
            slices_to_append.append(nn)

        slice_parent._movable = True # Force movable
        element._movable = True # Force movable

        return slices_to_append

    def _make_copies(self, element_names, index):
        new_names = []
        for element_name in element_names:
            element = self._line.element_dict[element_name]
            new_element_name = f'{element_name}..{index}'
            self._line.element_dict[new_element_name] = element.copy()
            new_names.append(new_element_name)

        return new_names

    def _order_set_by_line(self, set_to_order: set):
        """Order a set of element names by their order in the line."""
        assert isinstance(set_to_order, (set, frozenset))
        try:
            return sorted(set_to_order, key=self._line.element_names.index)
        except:
            return set_to_order



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib

import numpy as np
import pandas as pd
from cpymad.madx import Madx
from scipy.constants import c as clight

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_acceleration(test_context):
    Delta_p0c = 450e9/10*23e-6  # ramp rate 450GeV/10s

    fname_line = test_data_folder.joinpath(
        'sps_w_spacecharge/line_no_spacecharge_and_particle.json')

    with open(fname_line, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])

    energy_increase = xt.ReferenceEnergyIncrease(Delta_p0c=Delta_p0c)
    line.append_element(energy_increase, 'energy_increase')

    line.build_tracker(_context=test_context)

    # Assume only first cavity is active
    frequency = line.get_elements_of_type(xt.Cavity)[0][0].frequency
    voltage = line.get_elements_of_type(xt.Cavity)[0][0].voltage
    # Assuming proton and beta=1
    stable_z = np.arcsin(Delta_p0c/voltage)/frequency/2/np.pi*clight

    p_co = line.find_closed_orbit(particle_ref=xp.Particles.from_dict(
        input_data['particle']))

    xo.assert_allclose(p_co._xobject.zeta[0], stable_z, atol=0, rtol=1e-2)


@for_all_test_contexts(excluding=('ContextPyopencl',))
def test_energy_program(test_context):

    df = pd.read_csv(test_data_folder / 'psb_chicane/Ramp_and_RF_functions.dat',
                     sep='\t', skiprows=2,
                     names=['t_s', 'E_kin_GeV', 'V1_MV', 'phi1_rad', 'V2_MV', 'phi2_rad'])
    E_kin_GeV = df.E_kin_GeV.values
    t_s = df.t_s.values

    # Shift to enhance change in revolution frequency
    E_min = np.min(E_kin_GeV)
    E_max = np.max(E_kin_GeV)
    E_kin_GeV = E_min/100 + (E_kin_GeV - E_min)
    # Shift the time scale for testing purposes
    t_s = t_s

    # Load mad model and apply element shifts
    mad = Madx(stdout=False)
    mad.call(str(test_data_folder / 'psb_chicane/psb.seq'))
    mad.call(str(test_data_folder / 'psb_chicane/psb_fb_lhc.str'))
    mad.input('''
        beam, particle=PROTON, pc=0.5708301551893517;
        use, sequence=psb1;
        twiss;
    ''')

    line = xt.Line.from_madx_sequence(mad.sequence.psb1, allow_thick=True,
                                      deferred_expressions=True)
    line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV,
                                     gamma0=mad.sequence.psb1.beam.gamma)

    line.build_tracker(_context=test_context)

    # Attach energy program
    ep = xt.EnergyProgram(t_s=t_s, kinetic_energy0=E_kin_GeV*1e9)
    line.energy_program = ep

    line['br1.acwf7l1.1'].voltage = 2e6
    line['br1.acwf7l1.1'].frequency = 1e3

    # test to_dict and from_dict
    lbefore = line
    line_dict = line.to_dict()
    line = xt.Line.from_dict(line_dict)
    assert np.all(line.vars.get_table().name == lbefore.vars.get_table().name)

    line = line.copy()
    assert np.all(line.vars.get_table().name == lbefore.vars.get_table().name)

    line.build_tracker(_context=test_context)

    p_test = line.build_particles()
    line.enable_time_dependent_vars = True
    n_turn_test = 5000
    monitor = xt.ParticlesMonitor(num_particles=len(p_test.zeta), start_at_turn=0,
                                  stop_at_turn=n_turn_test,
                                  _context=test_context)
    for ii in range(n_turn_test):
        line.track(p_test, turn_by_turn_monitor=monitor)

    beta_at_turn = monitor.beta0[0, :]
    gamma_at_turn = 1 / np.sqrt(1 - beta_at_turn**2)

    t_turn_ref = np.cumsum(line.get_length()/clight/beta_at_turn)
    t_turn_ref = t_turn_ref - t_turn_ref[0]
    E_kin_turn = line.particle_ref.mass0 * (monitor.gamma0[0, :] - 1)

    t_check = np.linspace(0, 20e-3, 1000)
    E_check = np.interp(t_check, t_turn_ref, E_kin_turn)
    E_check_ref = np.interp(t_check, t_s, E_kin_GeV*1e9)
    xo.assert_allclose(E_check, E_check_ref, atol=0, rtol=2e-3)

    t_turn_check = line.energy_program.get_t_s_at_turn(np.arange(n_turn_test))
    xo.assert_allclose(t_turn_check, t_turn_ref, atol=0, rtol=6e-4)

    p0c_check = line.energy_program.get_p0c_at_t_s(t_check)
    p0c_ref = np.interp(t_check,
                        t_turn_check,
                        line.particle_ref.mass0 * gamma_at_turn * beta_at_turn)
    xo.assert_allclose(p0c_check, p0c_ref, atol=0, rtol=1e-3)

    kinetic_energy0_check = line.energy_program.get_kinetic_energy0_at_t_s(t_check)
    kinetic_energy0_ref = np.interp(t_check,
                        t_turn_check,
                        line.particle_ref.mass0 * (gamma_at_turn - 1))
    xo.assert_allclose(kinetic_energy0_check, kinetic_energy0_ref, atol=0, rtol=2e-3)

    beta0_check = line.energy_program.get_beta0_at_t_s(t_check)
    beta0_ref = np.interp(t_check, t_turn_check, beta_at_turn)
    xo.assert_allclose(beta0_check, beta0_ref, atol=0, rtol=1e-3)

    frev_check = line.energy_program.get_frev_at_t_s(t_check)
    frev_ref = np.interp(t_check, t_turn_check[:-1], 1/np.diff(t_turn_ref))
    xo.assert_allclose(frev_check, frev_ref, atol=0, rtol=4e-5)

    p0c_increse_per_turn_check = line.energy_program.get_p0c_increse_per_turn_at_t_s(
        t_check)
    p0c_increse_per_turn_ref = np.interp(
        t_check, t_turn_check[:-1], np.diff(monitor.p0c[0, :]))
    xo.assert_allclose(p0c_increse_per_turn_check - p0c_increse_per_turn_ref, 0,
                       atol=5e-5 * p0c_ref[0], rtol=0)

    line.enable_time_dependent_vars = False
    line.vars['t_turn_s'] = 20e-3

    E_kin_expected = np.interp(line.vv['t_turn_s'], t_s, E_kin_GeV*1e9)
    E_tot_expected = E_kin_expected + line.particle_ref.mass0
    xo.assert_allclose(
        E_tot_expected, line.particle_ref.energy0[0], rtol=1e-4, atol=0)
    xo.assert_allclose(
        E_kin_expected, line.particle_ref.kinetic_energy0[0], rtol=1e-4, atol=0)

    tw = line.twiss(method='6d')
    # To check that it does not change
    xo.assert_allclose(tw.zeta[0], -13.48, rtol=0, atol=1e-4)
    xo.assert_allclose(line.particle_ref.mass0 * tw.gamma0, E_tot_expected,
                      atol=0, rtol=1e-12)

    line.vars['t_turn_s'] = 0
    line.vars['on_chicane_k0'] = 0
    tw = line.twiss(method='6d')
    xo.assert_allclose(tw.zeta[0], 0, rtol=0, atol=1e-12)
    xo.assert_allclose(line.particle_ref.mass0 * tw.gamma0, line.particle_ref.mass0 + E_kin_turn[0],
                       rtol=1e-10, atol=0)

@for_all_test_contexts(excluding=('ContextPyopencl',))
def test_acceleration_transverse_shrink(test_context):

    mad = Madx(stdout=False)

    # Load mad model and apply element shifts
    mad.input(f'''
    call, file = '{str(test_data_folder)}/psb_chicane/psb.seq';
    call, file = '{str(test_data_folder)}/psb_chicane/psb_fb_lhc.str';
    beam;
    use, sequence=psb1;
    ''')

    line = xt.Line.from_madx_sequence(mad.sequence.psb1,
                                        deferred_expressions=True)
    e_kin_start_eV = 160e6
    line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1.,
                                    energy0=xt.PROTON_MASS_EV + e_kin_start_eV)

    # Slice to gain some tracking speed
    line.slice_thick_elements(
        slicing_strategies=[
            xt.Strategy(slicing=xt.Teapot(1)),
        ])
    line.build_tracker(_context=test_context)

    # User-defined energy ramp
    t_s = np.array([0., 0.0006, 0.0008, 0.001 , 0.0012, 0.0014, 0.0016, 0.0018,
                    0.002 , 0.0022, 0.0024, 0.0026, 0.0028, 0.003, 0.01, 0.1])
    E_kin_GeV = np.array([0.16000000,0.16000000,
        0.16000437, 0.16001673, 0.16003748, 0.16006596, 0.16010243, 0.16014637,
        0.16019791, 0.16025666, 0.16032262, 0.16039552, 0.16047524, 0.16056165,
        0.163586, 0.20247050000000014])

    # Enhance energy swing to better see the effect of energy on beam size
    E_kin_GeV -= 0.140

    # Go away from half integer
    opt = line.match(
        #verbose=True,
        method='4d',
        solve=False,
        vary=[
            xt.Vary('kbrqfcorr', step=1e-4),
            xt.Vary('kbrqdcorr', step=1e-4),
        ],
        targets = [
            xt.Target('qx', value=4.15, tol=1e-5, scale=1),
            xt.Target('qy', value=4.18, tol=1e-5, scale=1),
        ]
    )
    opt.solve()


    # Attach energy program to the line
    line.energy_program = xt.EnergyProgram(t_s=t_s, kinetic_energy0=E_kin_GeV*1e9)

    # Plot energy and revolution frequency vs time
    t_plot = np.linspace(0, 10e-3, 20)
    E_kin_plot = line.energy_program.get_kinetic_energy0_at_t_s(t_plot)
    f_rev_plot = line.energy_program.get_frev_at_t_s(t_plot)

    # import matplotlib.pyplot as plt
    # plt.close('all')
    # plt.figure(1, figsize=(6.4 * 1.5, 4.8))
    # ax1 = plt.subplot(2,2,1)
    # plt.plot(t_plot * 1e3, E_kin_plot * 1e-6)
    # plt.ylabel(r'$E_{kin}$ [MeV]')
    # ax2 = plt.subplot(2,2,3, sharex=ax1)
    # plt.plot(t_plot * 1e3, f_rev_plot * 1e-3)
    # plt.ylabel(r'$f_{rev}$ [kHz]')
    # plt.xlabel('t [ms]')

    # Setup frequency of the RF cavity  to stay on the second harmonic of the
    # revolution frequency during the acceleration

    t_rf = np.linspace(0, 3e-3, 100) # time samples for the frequency program
    f_rev = line.energy_program.get_frev_at_t_s(t_rf)
    h_rf = 2 # harmonic number
    f_rf = h_rf * f_rev # frequency program

    # Build a function with these samples and link it to the cavity
    line.functions['fun_f_rf'] = xt.FunctionPieceWiseLinear(x=t_rf, y=f_rf)
    line.element_refs['br1.acwf5l1.1'].frequency = line.functions['fun_f_rf'](
                                                            line.vars['t_turn_s'])

    # Setup voltage and lag
    line.element_refs['br1.acwf5l1.1'].voltage = 3000 # V
    line.element_refs['br1.acwf5l1.1'].lag = 0 # degrees (below transition energy)

    # When setting line.vars['t_turn_s'] the reference energy and the rf frequency
    # are updated automatically
    line.vars['t_turn_s'] = 0
    line.particle_ref.kinetic_energy0 # is 160.00000 MeV
    line['br1.acwf5l1.1'].frequency # is 1983931.935 Hz

    line.vars['t_turn_s'] = 3e-3
    line.particle_ref.kinetic_energy0 # is 160.56165 MeV
    line['br1.acwf5l1.1'].frequency # is 1986669.0559674294

    # Back to zero for tracking!
    line.vars['t_turn_s'] = 0

    # Track a few particles to visualize the longitudinal phase space
    p_test = line.build_particles(x_norm=0, zeta=np.linspace(0, line.get_length(), 101))

    # Enable time-dependent variables (t_turn_s and all the variables that depend on
    # it are automatically updated at each turn)
    line.enable_time_dependent_vars = True

    # Track
    line.track(p_test, num_turns=9000, turn_by_turn_monitor=True, with_progress=True)
    mon = line.record_last_track

    # Plot
    # plt.subplot2grid((2,2), (0,1), rowspan=2)
    # plt.plot(mon.zeta[:, -2000:].T, mon.delta[:, -2000:].T, color='C0')
    # plt.xlabel(r'$\zeta$ [m]')
    # plt.ylabel('$\delta$')
    # plt.xlim(-40, 30)
    # plt.ylim(-0.0025, 0.0025)
    # plt.title('Last 2000 turns')
    # plt.subplots_adjust(left=0.08, right=0.95, wspace=0.26)


    # Check transverse beam size reduction
    line['t_turn_s'] = 0
    line.enable_time_dependent_vars = False

    n_part_test = 500
    # Generate Gaussian distribution with fixed rng seed
    rng = np.random.default_rng(seed=123)
    x_norm = rng.normal(loc=0, scale=1, size=n_part_test)
    px_norm = rng.normal(loc=0, scale=1, size=n_part_test)
    y_norm = rng.normal(loc=0, scale=1, size=n_part_test)
    py_norm = rng.normal(loc=0, scale=1, size=n_part_test)

    # rescale to have exact std dev.
    x_norm = x_norm / np.std(x_norm)
    px_norm = px_norm / np.std(px_norm)
    y_norm = y_norm / np.std(y_norm)
    py_norm = py_norm / np.std(py_norm)

    p_test2 = line.build_particles(x_norm=x_norm, px_norm=px_norm,
                                y_norm=x_norm, py_norm=px_norm,
                                nemitt_x=1e-6, nemitt_y=1e-6,
                                delta=0)

    line.enable_time_dependent_vars = True
    line.track(p_test2, num_turns=10_000, turn_by_turn_monitor=True, with_progress=True)
    mon2 = line.record_last_track

    std_y = np.std(mon2.y, axis=0)
    std_x = np.std(mon2.x, axis=0)

    # Apply moving average filter
    from scipy.signal import savgol_filter
    std_y_smooth = savgol_filter(std_y, 10000, 2)
    std_x_smooth = savgol_filter(std_x, 10000, 2)

    i_turn_match = 1000
    std_y_expected = std_y_smooth[i_turn_match] * np.sqrt(
        mon2.gamma0[0, i_turn_match]* mon2.beta0[0, i_turn_match]
        / mon2.gamma0[0, :] / mon2.beta0[0, :])
    std_x_expected = std_x_smooth[i_turn_match] * np.sqrt(
        mon2.gamma0[0, i_turn_match]* mon2.beta0[0, i_turn_match]
        / mon2.gamma0[0, :] / mon2.beta0[0, :])

    d_sigma_x = std_x_expected[0] - std_x_expected[-1]
    d_sigma_y = std_y_expected[0] - std_y_expected[-1]

    import xobjects as xo
    xo.assert_allclose(std_y_expected[8000:9000].mean(),
                    std_y_smooth[8000:9000].mean(),
                    rtol=0, atol=0.03 * d_sigma_y)
    xo.assert_allclose(std_x_expected[8000:9000].mean(),
                    std_x_smooth[8000:9000].mean(),
                    rtol=0, atol=0.03 * d_sigma_x)

    # plt.figure(2)
    # ax1 = plt.subplot(2,1,1)
    # plt.plot(std_x, label='raw')
    # plt.plot(std_x_smooth, label='smooth')
    # plt.plot(std_x_expected, label='expected')
    # plt.legend()

    # ax2 = plt.subplot(2,1,2, sharex=ax1)
    # plt.plot(std_y, label='raw')
    # plt.plot(std_y_smooth, label='smooth')
    # plt.plot(std_y_expected, label='expected')
    # plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import logging

import numpy as np
import pytest

import xobjects as xo
import xpart as xp
import xtrack as xt


def test_collimation_infrastructure():

    ######################################
    # Create a dummy collimation process #
    ######################################

    class DummyInteractionProcess:

        def __init__(self, fraction_lost, fraction_secondary):

            self.fraction_lost = fraction_lost
            self.fraction_secondary = fraction_secondary

        def interact(self, particles):

            n_part = particles._num_active_particles

            # Kill some particles
            mask_kill = np.random.uniform(size=n_part) < self.fraction_lost
            particles.state[:n_part][mask_kill] = 0

            # Generate some more particles
            mask_secondary = np.random.uniform(size=n_part) < self.fraction_secondary
            n_products = np.sum(mask_secondary)
            if n_products>0:
                products = {
                    's': particles.s[:n_part][mask_secondary],
                    'x': particles.x[:n_part][mask_secondary],
                    'px': particles.px[:n_part][mask_secondary],
                    'y': particles.y[:n_part][mask_secondary],
                    'py': particles.py[:n_part][mask_secondary],
                    'zeta': particles.zeta[:n_part][mask_secondary],
                    'delta': particles.delta[:n_part][mask_secondary],

                    'mass_ratio': particles.x[:n_part][mask_secondary] *0 + .5,
                    'charge_ratio': particles.x[:n_part][mask_secondary] *0 + .5,

                    'parent_particle_id': particles.particle_id[:n_part][
                                                                mask_secondary],
                    'at_element': particles.at_element[:n_part][mask_secondary],
                    'at_turn': particles.at_turn[:n_part][mask_secondary],
                    }
            else:
                products = None

            return products

    #############################################
    # Create the corresponding beam interaction #
    #############################################

    beam_interaction = xt.BeamInteraction(
            interaction_process=DummyInteractionProcess(fraction_lost=0.1,
                                                        fraction_secondary=0.2))

    ############################################
    # Go through the collimator multiple times #
    ############################################

    particles = xp.Particles(_capacity=200,
            p0c=7000, x=np.linspace(-1e-3, 1e-3, 10))

    for _ in range(10):
        beam_interaction.track(particles)

    ###############
    # Some checks #
    ###############

    assert particles._num_lost_particles >= 0
    assert particles._num_active_particles >= 0

    n_all_parts = particles._num_active_particles + particles._num_lost_particles

    assert np.all(np.diff(particles.state) <= 0) # checks that there is no lost after active

    # Check each id is present only once
    ids = particles.particle_id[:n_all_parts]
    assert len(list(set(ids))) == n_all_parts

    # Check parent and secondaries have the same position
    ind_secondaries = np.where(particles.parent_particle_id[:particles._num_active_particles] !=
                        particles.particle_id[:particles._num_active_particles])[0]
    for ii in ind_secondaries:
        parent_id = particles.parent_particle_id[ii]
        parent_x = particles.x[np.where(particles.particle_id == parent_id)[0][0]]

        assert parent_x == particles.x[ii]

@pytest.mark.parametrize(
    'sandwitch_aper',
    [True, False],
    ids=['sadwitch_aper', 'no_sadwitch_aper'])
def test_aperture_refinement(sandwitch_aper):
    n_part=10000
    shift_x = 0.3e-2
    shift_y = 0.5e-2

    ctx = xo.context_default
    buf = ctx.new_buffer()

    logger = logging.getLogger('xtrack')
    logger.setLevel(logging.DEBUG)

    # Define aper_0
    aper_0 = xt.LimitEllipse(_buffer=buf, a=2e-2, b=2e-2)
    shift_aper_0 = (shift_x, shift_y)
    rot_deg_aper_0 = 10.

    # Define aper_1
    aper_1 = xt.LimitEllipse(_buffer=buf, a=1e-2, b=1e-2)
    shift_aper_1 = (shift_x, shift_y)
    rot_deg_aper_1 = 10.

    # aper_0_sandwitch
    if sandwitch_aper:
        line_aper_0 = xt.Line(
            elements=[xt.XYShift(_buffer=buf, dx=shift_aper_0[0], dy=shift_aper_0[1]),
                    xt.SRotation(_buffer=buf, angle=rot_deg_aper_0),
                    aper_0,
                    xt.Multipole(_buffer=buf, knl=[0.00]),
                    xt.SRotation(_buffer=buf, angle=-rot_deg_aper_0),
                    xt.XYShift(_buffer=buf, dx=-shift_aper_0[0], dy=-shift_aper_0[1])])
        line_aper_0.build_tracker(_buffer=buf)
        # aper_1_sandwitch
        line_aper_1 = xt.Line(
            elements=[xt.XYShift(_buffer=buf, dx=shift_aper_1[0], dy=shift_aper_1[1]),
                    xt.SRotation(_buffer=buf, angle=rot_deg_aper_1),
                    aper_1,
                    xt.Multipole(_buffer=buf, knl=[0.00]),
                    xt.SRotation(_buffer=buf, angle=-rot_deg_aper_1),
                    xt.XYShift(_buffer=buf, dx=-shift_aper_1[0], dy=-shift_aper_1[1])])
    else:
        aper_0.shift_x = shift_aper_0[0]
        aper_0.shift_y = shift_aper_0[1]
        aper_0.rot_s_rad = np.deg2rad(rot_deg_aper_0)
        line_aper_0 = xt.Line(
            elements=[aper_0, xt.Multipole(_buffer=buf, knl=[0.0])])
        line_aper_0.build_tracker(_buffer=buf)
        aper_1.shift_x = shift_aper_1[0]
        aper_1.shift_y = shift_aper_1[1]
        aper_1.rot_s_rad = np.deg2rad(rot_deg_aper_1)
        line_aper_1 = xt.Line(
            elements=[aper_1, xt.Multipole(_buffer=buf, knl=[0.00])])

    line_aper_1.build_tracker(_buffer=buf)

    # Build example line
    line=xt.Line(
        elements = ((xt.Drift(_buffer=buf, length=0.5),)
                    + line_aper_0.elements
                    + (xt.Drift(_buffer=buf, length=1),
                       xt.Multipole(_buffer=buf, knl=[0.]),
                       xt.Drift(_buffer=buf, length=1),
                       xt.Cavity(_buffer=buf, voltage=3e6, frequency=400e6),
                       xt.ParticlesMonitor(_buffer=buf,
                            start_at_turn=0, stop_at_turn=10, num_particles=3),
                       xt.Drift(_buffer=buf, length=1.),
                       xt.Marker())
                    + line_aper_1.elements))
    line.build_tracker(_buffer=buf)

    # Test on full line
    r = np.linspace(0, 0.018, n_part)
    theta = np.linspace(0, 8*np.pi, n_part)
    particles = xp.Particles(_context=ctx,
            p0c=6500e9,
            x=r*np.cos(theta)+shift_x,
            y=r*np.sin(theta)+shift_y)

    line.track(particles)


    loss_loc_refinement = xt.LossLocationRefinement(line,
                            n_theta = 360,
                            r_max = 0.5, # m
                            dr = 50e-6,
                            ds = 0.1,
                            save_refine_lines=True,
                            allowed_backtrack_types=[
                                xt.Multipole,
                                xt.Cavity
                                ])

    import time
    t0 = time.time()

    loss_loc_refinement.refine_loss_location(particles)

    t1 = time.time()
    print(f'Took\t{(t1-t0)*1e3:.2f} ms')

    # Automatic checks
    mask_lost = particles.state == 0
    r_calc = np.sqrt((particles.x-shift_x)**2 + (particles.y-shift_y)**2)
    assert np.all(r_calc[~mask_lost]<1e-2)
    assert np.all(r_calc[mask_lost]>1e-2)
    i_aper_1 = line.elements.index(aper_1)
    assert np.all(particles.at_element[mask_lost]==i_aper_1)
    assert np.all(particles.at_element[~mask_lost]==0)
    s0 = line.get_s_elements()[line.elements.index(aper_0)]
    s1 = line.get_s_elements()[line.elements.index(aper_1)]
    r0 = np.sqrt(aper_0.a_squ)
    r1 = np.sqrt(aper_1.a_squ)
    s_expected = s0 + (r_calc-r0)/(r1 - r0)*(s1 - s0)
    # TODO This threshold is a bit large
    xo.assert_allclose(particles.s[mask_lost], s_expected[mask_lost], atol=0.11)


def test_losslocationrefinement_thick_collective_collimator():

    ######################################
    # Create a dummy collimation process #
    ######################################

    class DummyInteractionProcess:
        '''
        I kill some particles and I kick some others by an given angle
        and I generate some secondaries with the opposite angles.
        '''
        def __init__(self, fraction_lost, fraction_secondary, length, kick_x):

            self.fraction_lost = fraction_lost
            self.fraction_secondary = fraction_secondary
            self.kick_x = kick_x
            self.length = length

            self.drift= xt.Drift(length=self.length)


        def interact(self, particles):

            self.drift.track(particles)

            n_part = particles._num_active_particles

            # Kill some particles
            mask_kill = np.random.uniform(size=n_part) < self.fraction_lost
            particles.state[:n_part][mask_kill] = -1 # special flag`


            # Generate some more particles
            mask_secondary = np.random.uniform(size=n_part) < self.fraction_secondary
            n_products = np.sum(mask_secondary)
            if n_products>0:
                products = {
                    's': particles.s[:n_part][mask_secondary],
                    'x': particles.x[:n_part][mask_secondary],
                    'px': particles.px[:n_part][mask_secondary] + self.kick_x,
                    'y': particles.y[:n_part][mask_secondary],
                    'py': particles.py[:n_part][mask_secondary],
                    'zeta': particles.zeta[:n_part][mask_secondary],
                    'delta': particles.delta[:n_part][mask_secondary],

                    'mass_ratio': particles.x[:n_part][mask_secondary] *0 + 1.,
                    'charge_ratio': particles.x[:n_part][mask_secondary] *0 + 1.,

                    'parent_particle_id': particles.particle_id[:n_part][mask_secondary],
                    'at_element': particles.at_element[:n_part][mask_secondary],
                    'at_turn': particles.at_turn[:n_part][mask_secondary],
                    }
            else:
                products = None

            return products

    #############################################
    # Create the corresponding beam interaction #
    #############################################

    interaction_process=DummyInteractionProcess(length=1., kick_x=4e-3,
                                                fraction_lost=0.0,
                                                fraction_secondary=0.2)
    beam_interaction = xt.BeamInteraction(length=interaction_process.length,
                                          interaction_process=interaction_process)

    line = xt.Line(elements=[
        xt.Multipole(knl=[0,0]),
        xt.LimitEllipse(a=2e-2, b=2e-2),
        xt.Drift(length=1.),
        xt.Multipole(knl=[0,0]),
        xt.LimitEllipse(a=2e-2, b=2e-2),
        xt.Drift(length=2.),
        beam_interaction,
        xt.Multipole(knl=[0,0]),
        xt.LimitEllipse(a=2e-2, b=2e-2),
        xt.Drift(length=10.),
        xt.LimitEllipse(a=2e-2, b=2e-2),
        xt.Drift(length=10.),
        ])
    line.build_tracker()

    particles = xp.Particles(
            _capacity=200000,
            x=np.zeros(100000))

    line.track(particles)

    mask_lost = particles.state == 0
    assert np.all(particles.at_element[mask_lost] == 10)


    loss_loc_refinement = xt.LossLocationRefinement(line,
                                                n_theta = 360,
                                                r_max = 0.5, # m
                                                dr = 50e-6,
                                                ds = 0.05,
                                                save_refine_lines=True)

    loss_loc_refinement.refine_loss_location(particles)

    xo.assert_allclose(particles.s[mask_lost], 9.00,
                       rtol=0, atol=loss_loc_refinement.ds*1.0001)

def test_losslocationrefinement_skip_refinement_for_collimators():

    class DummyInteractionProcess:
        '''
        I kill some particles. I kick some others by an given angle
        and I generate some secondaries with the opposite angles.
        '''
        def __init__(self, fraction_lost, fraction_secondary, length, kick_x):

            self.fraction_lost = fraction_lost
            self.fraction_secondary = fraction_secondary
            self.kick_x = kick_x
            self.length = length

            self.drift = xt.Drift(length=self.length)


        def interact(self, particles):

            self.drift.track(particles)

            n_part = particles._num_active_particles

            # Kill some particles
            mask_kill = np.random.uniform(size=n_part) < self.fraction_lost
            particles.state[:n_part][mask_kill] = -1 # special flag`


            # Generate some more particles
            mask_secondary = np.random.uniform(size=n_part) < self.fraction_secondary
            n_products = np.sum(mask_secondary)
            if n_products>0:
                products = {
                    's': particles.s[:n_part][mask_secondary],
                    'x': particles.x[:n_part][mask_secondary] + self.length * (
                        particles.px[:n_part][mask_secondary] + self.kick_x),
                    'px': particles.px[:n_part][mask_secondary] + self.kick_x,
                    'y': particles.y[:n_part][mask_secondary],
                    'py': particles.py[:n_part][mask_secondary],
                    'zeta': particles.zeta[:n_part][mask_secondary],
                    'delta': particles.delta[:n_part][mask_secondary],

                    'mass_ratio': particles.x[:n_part][mask_secondary] *0 + 1.,
                    'charge_ratio': particles.x[:n_part][mask_secondary] *0 + 1.,

                    'parent_particle_id': particles.particle_id[:n_part][mask_secondary],
                    'at_element': particles.at_element[:n_part][mask_secondary],
                    'at_turn': particles.at_turn[:n_part][mask_secondary],
                    }
            else:
                products = None

            return products

    ############################################################
    # Create a beam interaction from the process defined above #
    ############################################################

    # kick_x is large so that the particles are lost on the first aperture element
    interaction_process=DummyInteractionProcess(length=1., kick_x=4e-2,
                                                fraction_lost=0.0,
                                                fraction_secondary=0.2)
    collimator = xt.BeamInteraction(length=interaction_process.length,
                                        interaction_process=interaction_process)

    ########################################################
    # Create a line including the collimator defined above #
    ########################################################

    line = xt.Line(elements=[
        xt.Drift(length=2.),
        xt.Marker(),
        xt.LimitEllipse(a=2e-2, b=2e-2),
        collimator,
        xt.Marker(),
        xt.LimitEllipse(a=2e-2, b=2e-2),
        xt.Drift(length=10.),
        ])

    #################
    # Build line #
    #################

    line.build_tracker()
    line.config.XTRACK_GLOBAL_XY_LIMIT = 1e3

    ##########################
    # Build particles object #
    ##########################

    # We prepare empty slots to store the product particles that will be
    # generated during the tracking.
    particles = xp.Particles(_capacity=200000, x=np.zeros(100000))

    #########
    # Track #
    #########

    line.track(particles)

    ############################
    # Loss location refinement #
    ############################

    part_before = particles.copy()

    loss_loc_refinement = xt.LossLocationRefinement(line,
                                                n_theta = 360,
                                                r_max = 0.5, # m
                                                dr = 50e-6,
                                                ds = 0.05,
                                                save_refine_lines=True)

    loss_loc_refinement.refine_loss_location(particles)

    assert np.all(part_before.s == particles.s)



import numpy as np

import xdeps as xd
import xtrack as xt

# Values to check
d_mux_15_b1 = 0.07
d_muy_15_b1 = 0.05
d_mux_15_b2 = -0.1
d_muy_15_b2 = -0.12

default_tol = {None: 1e-8, 'betx': 1e-6, 'bety': 1e-6} # to have no rematching w.r.t. madx

collider = xt.Environment.from_json('collider_01_changed_ip15_phase.json')
collider.build_trackers()

collider_ref = xt.Environment.from_json('collider_00_from_madx.json')
collider_ref.build_trackers()
collider_ref.vars.load_madx_optics_file(
    "../../test_data/hllhc15_thick/opt_round_150_1500.madx")

tw = collider.twiss()
tw_ref = collider_ref.twiss()

d_mux = {'lhcb1': d_mux_15_b1, 'lhcb2': d_mux_15_b2}
d_muy = {'lhcb1': d_muy_15_b1, 'lhcb2': d_muy_15_b2}

for ll in ['lhcb1', 'lhcb2']:

    assert np.isclose(tw[ll].qx, tw_ref[ll].qx, atol=1e-7, rtol=0)
    assert np.isclose(tw[ll].qy, tw_ref[ll].qy, atol=1e-7, rtol=0)

    mux_15 = tw[ll]['mux', 'ip5'] - tw[ll]['mux', 'ip1']
    mux_ref_15 = tw_ref[ll]['mux', 'ip5'] - tw_ref[ll]['mux', 'ip1']

    muy_15 = tw[ll]['muy', 'ip5'] - tw[ll]['muy', 'ip1']
    muy_ref_15 = tw_ref[ll]['muy', 'ip5'] - tw_ref[ll]['muy', 'ip1']

    assert np.isclose(mux_15, mux_ref_15 + d_mux[ll], atol=1e-7, rtol=0)
    assert np.isclose(muy_15, muy_ref_15 + d_muy[ll], atol=1e-7, rtol=0)

    twip = tw[ll].rows['ip.*']
    twip_ref = tw_ref[ll].rows['ip.*']

    assert np.allclose(twip['betx'], twip_ref['betx'], rtol=1e-6, atol=0)
    assert np.allclose(twip['bety'], twip_ref['bety'], rtol=1e-6, atol=0)
    assert np.allclose(twip['alfx'], twip_ref['alfx'], rtol=0, atol=1e-5)
    assert np.allclose(twip['alfy'], twip_ref['alfy'], rtol=0, atol=1e-5)
    assert np.allclose(twip['dx'], twip_ref['dx'], rtol=0, atol=1e-6)
    assert np.allclose(twip['dy'], twip_ref['dy'], rtol=0, atol=1e-6)
    assert np.allclose(twip['dpx'], twip_ref['dpx'], rtol=0, atol=1e-7)
    assert np.allclose(twip['dpy'], twip_ref['dpy'], rtol=0, atol=1e-7)


# -----------------------------------------------------------------------------

# Check higher level knobs

collider.vars['on_x2'] = 34
tw = collider.twiss()
collider.vars['on_x2'] = 0

assert np.isclose(tw.lhcb1['py', 'ip2'], 34e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip2'], -34e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-9, rtol=0)

collider.vars['on_x8'] = 35
tw = collider.twiss()
collider.vars['on_x8'] = 0

assert np.isclose(tw.lhcb1['px', 'ip8'], 35e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip8'], -35e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-9, rtol=0)

collider.vars['on_sep2'] = 0.5
tw = collider.twiss()
collider.vars['on_sep2'] = 0

assert np.isclose(tw.lhcb1['x', 'ip2'], -0.5e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip2'], 0.5e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-9, rtol=0)

collider.vars['on_sep8'] = 0.6
tw = collider.twiss()
collider.vars['on_sep8'] = 0

assert np.isclose(tw.lhcb1['y', 'ip8'], 0.6e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip8'], -0.6e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-9, rtol=0)

# Check lower level knobs (disconnects higher level knobs)

collider.vars['on_o2v'] = 0.3
tw = collider.twiss()
collider.vars['on_o2v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip2'], 0.3e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip2'], 0.3e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip2'], 0., atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip2'], 0., atol=1e-9, rtol=0)

collider.vars['on_o2h'] = 0.4
tw = collider.twiss()
collider.vars['on_o2h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip2'], 0.4e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip2'], 0.4e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip2'], 0., atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip2'], 0., atol=1e-9, rtol=0)

collider.vars['on_o8v'] = 0.5
tw = collider.twiss()
collider.vars['on_o8v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip8'], 0.5e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip8'], 0.5e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip8'], 0., atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip8'], 0., atol=1e-9, rtol=0)

collider.vars['on_o8h'] = 0.6
tw = collider.twiss()
collider.vars['on_o8h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip8'], 0.6e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], 0.6e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip8'], 0., atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip8'], 0., atol=1e-9, rtol=0)

collider.vars['on_a2h'] = 20
tw = collider.twiss()
collider.vars['on_a2h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip2'], 20e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip2'], 20e-6, atol=1e-9, rtol=0)

collider.vars['on_a2v'] = 15
tw = collider.twiss()
collider.vars['on_a2v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip2'], 15e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip2'], 15e-6, atol=1e-9, rtol=0)

collider.vars['on_a8h'] = 20
tw = collider.twiss()
collider.vars['on_a8h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip8'], 20e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip8'], 20e-6, atol=1e-9, rtol=0)

collider.vars['on_a8v'] = 50
tw = collider.twiss()
collider.vars['on_a8v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip8'], 50e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip8'], 50e-6, atol=1e-9, rtol=0)

collider.vars['on_x2v'] = 100
tw = collider.twiss()
collider.vars['on_x2v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip2'], 100e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip2'], -100e-6, atol=1e-9, rtol=0)

collider.vars['on_x2h'] = 120
tw = collider.twiss()
collider.vars['on_x2h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip2'], 120e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip2'], -120e-6, atol=1e-9, rtol=0)


collider.vars['on_x8h'] = 100
tw = collider.twiss()
collider.vars['on_x8h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip8'], 100e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip8'], -100e-6, atol=1e-9, rtol=0)

collider.vars['on_x8v'] = 120
tw = collider.twiss()
collider.vars['on_x8v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip8'], 120e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip8'], -120e-6, atol=1e-9, rtol=0)

collider.vars['on_sep2h'] = 1.6
tw = collider.twiss()
collider.vars['on_sep2h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip2'], 1.6e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip2'], -1.6e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-9, rtol=0)

collider.vars['on_sep2v'] = 1.7
tw = collider.twiss()
collider.vars['on_sep2v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip2'], 1.7e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip2'], -1.7e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip2'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip2'], 0, atol=1e-9, rtol=0)

collider.vars['on_sep8h'] = 1.5
tw = collider.twiss()
collider.vars['on_sep8h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip8'], 1.5e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], -1.5e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip8'], 0, atol=1e-9, rtol=0)

collider.vars['on_sep8v'] = 1.7
tw = collider.twiss()
collider.vars['on_sep8v'] = 0

assert np.isclose(tw.lhcb1['y', 'ip8'], 1.7e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['y', 'ip8'], -1.7e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-9, rtol=0)

# Both knobs together
collider.vars['on_x8h'] = 120
collider.vars['on_sep8h'] = 1.7
tw = collider.twiss()
collider.vars['on_x8h'] = 0
collider.vars['on_sep8h'] = 0

assert np.isclose(tw.lhcb1['x', 'ip8'], 1.7e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['x', 'ip8'], -1.7e-3, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb1['px', 'ip8'], 120e-6, atol=1e-9, rtol=0)
assert np.isclose(tw.lhcb2['px', 'ip8'], -120e-6, atol=1e-9, rtol=0)


# Load generated optics file in MAD-X
from cpymad.madx import Madx

mad=Madx()
mad.call('../../test_data/hllhc15_thick/lhc.seq')
mad.call('../../test_data/hllhc15_thick/hllhc_sequence.madx')
mad.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
mad.use('lhcb1')
mad.input('beam, sequence=lhcb2, particle=proton, energy=7000, bv=-1;')
mad.use('lhcb2')
mad.call("opt_round_150_1500_xs.madx")

mad.input('twiss, sequence=lhcb1, table=twb1')
mad.input('twiss, sequence=lhcb2, table=twb2')
twmad_b1 = xd.Table(mad.table.twb1)
twmad_b2 = xd.Table(mad.table.twb2)

assert np.isclose(twmad_b1['betx', 'ip1:1'], 0.15, rtol=1e-8, atol=0)
assert np.isclose(twmad_b1['bety', 'ip1:1'], 0.15, rtol=1e-8, atol=0)
assert np.isclose(twmad_b2['betx', 'ip1:1'], 0.15, rtol=1e-8, atol=0)
assert np.isclose(twmad_b2['bety', 'ip1:1'], 0.15, rtol=1e-8, atol=0)

import xdeps as xd

twmad_b1.rows['ip.*'].cols['betx bety x y px py'].show()
twmad_b2.rows['ip.*'].cols['betx bety x y px py'].show()

# Test orbit knobs
mad.globals.on_x8 = 100
mad.globals.on_x2 = 110

mad.input('twiss, sequence=lhcb1, table=twb1')
mad.input('twiss, sequence=lhcb2, table=twb2')
twmad_b1 = xd.Table(mad.table.twb1)
twmad_b2 = xd.Table(mad.table.twb2)

assert np.isclose(twmad_b1['px', 'ip8:1'], 100e-6, rtol=0, atol=1e-10)
assert np.isclose(twmad_b2['px', 'ip8:1'], -100e-6, rtol=0, atol=1e-10)
assert np.isclose(twmad_b1['py', 'ip2:1'], 110e-6, rtol=0, atol=1e-10)
assert np.isclose(twmad_b2['py', 'ip2:1'], -110e-6, rtol=0, atol=1e-10)

# Match tunes and chromaticity in the Xsuite model
opt = collider.match(
    solve=False,
    vary=[
        xt.VaryList(['kqtf.b1', 'kqtd.b1', 'ksf.b1', 'ksd.b1'], step=1e-7),
        xt.VaryList(['kqtf.b2', 'kqtd.b2', 'ksf.b2', 'ksd.b2'], step=1e-7)],
    targets = [
        xt.TargetSet(line='lhcb1', qx=62.315, qy=60.325, tol=1e-10),
        xt.TargetSet(line='lhcb1', dqx=10.0, dqy=12.0, tol=1e-5),
        xt.TargetSet(line='lhcb2', qx=62.316, qy=60.324, tol=1e-10),
        xt.TargetSet(line='lhcb2', dqx=9.0, dqy=11.0, tol=1e-5)])
opt.solve()

# Transfer knobs to madx model and check matched values

for kk, vv in opt.get_knob_values().items():
    mad.globals[kk] = vv

mad.input('twiss, sequence=lhcb1, table=twb1')
mad.input('twiss, sequence=lhcb2, table=twb2')

assert np.isclose(mad.table.twb1.summary.q1, 62.315, rtol=0, atol=1e-6)
assert np.isclose(mad.table.twb1.summary.q2, 60.325, rtol=0, atol=1e-6)
assert np.isclose(mad.table.twb2.summary.q1, 62.316, rtol=0, atol=1e-6)
assert np.isclose(mad.table.twb2.summary.q2, 60.324, rtol=0, atol=1e-6)
assert np.isclose(mad.table.twb1.summary.dq1, 10.0, rtol=0, atol=0.3)
assert np.isclose(mad.table.twb1.summary.dq2, 12.0, rtol=0, atol=0.3)
assert np.isclose(mad.table.twb2.summary.dq1, 9.0, rtol=0, atol=0.3)
assert np.isclose(mad.table.twb2.summary.dq2, 11.0, rtol=0, atol=0.3)




import numpy as np
import xtrack as xt

bend = xt.Bend(k0=0.4, h=0.3, length=1)

line = xt.Line(
    elements=[bend, xt.Replica(parent_name='e0'), xt.Replica(parent_name='e1'),
                         xt.Drift(length=1.), xt.Replica(parent_name='e3'),
                         xt.Replica(parent_name='e4')])
length = line.get_length()

tt = line.get_table()

assert np.all(tt['name'][:-1] == ['e0', 'e1', 'e2', 'e3', 'e4', 'e5'])

_is_drift = xt.line._is_drift
_behaves_like_drift = xt.line._behaves_like_drift

assert not _is_drift(line['e0'], line)
assert not _is_drift(line['e1'], line)
assert not _is_drift(line['e2'], line)
assert _is_drift(line['e3'], line)
assert _is_drift(line['e4'], line)
assert _is_drift(line['e5'], line)

assert not _behaves_like_drift(line['e0'], line)
assert not _behaves_like_drift(line['e1'], line)
assert not _behaves_like_drift(line['e2'], line)
assert _behaves_like_drift(line['e3'], line)
assert _behaves_like_drift(line['e4'], line)
assert _behaves_like_drift(line['e5'], line)

line.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(xt.Uniform(3), name='e0'),
        xt.Strategy(xt.Uniform(3, mode='thick'), name='e1'),
        xt.Strategy(xt.Uniform(3), name='e2')])
assert_allclose = np.testing.assert_allclose
assert_allclose(length, line.get_length(), rtol=0, atol=1e-15)

assert np.all(line.get_table().name == [
       'e0_entry', 'e0..entry_map', 'drift_e0..0', 'e0..0', 'drift_e0..1',
       'e0..1', 'drift_e0..2', 'e0..2', 'drift_e0..3', 'e0..exit_map',
       'e0_exit', 'e1_entry', 'e1..entry_map', 'e1..0', 'e1..1', 'e1..2',
       'e1..exit_map', 'e1_exit', 'e2_entry', 'e2..entry_map',
       'drift_e2..0', 'e2..0', 'drift_e2..1', 'e2..1', 'drift_e2..2',
       'e2..2', 'drift_e2..3', 'e2..exit_map', 'e2_exit', 'e3', 'e4',
       'e5', '_end_point'])

for nn, ee in line.items():
    if nn.startswith('drift'):
        assert _is_drift(ee, line)
        assert _behaves_like_drift(ee, line)
    elif nn in ['e3', 'e4', 'e5']:
        assert _is_drift(ee, line)
        assert _behaves_like_drift(ee, line)
    elif nn.endswith('_entry') or nn.endswith('_exit'): # Markers
        assert not _is_drift(ee, line)
        assert _behaves_like_drift(ee, line)
    else:
        assert not _is_drift(ee, line)
        assert not _behaves_like_drift(ee, line)

line.cut_at_s([0.6, 1.5])
assert_allclose(length, line.get_length(), rtol=0, atol=1e-15)
assert_allclose(line.get_s_position('drift_e0..2..1'), 0.6, rtol=0, atol=1e-15)
assert_allclose(line.get_s_position('e1..1..1'), 1.5, rtol=0, atol=1e-15)

assert _is_drift(line['drift_e0..2..1'], line)
assert _behaves_like_drift(line['drift_e0..2..1'], line)
assert not _is_drift(line['e1..1..1'], line)
assert not _behaves_like_drift(line['e1..1..1'], line)

line.slice_thick_elements(
    slicing_strategies=[
        xt.Strategy(None),
        xt.Strategy(xt.Uniform(2), name='e3'),
        xt.Strategy(xt.Uniform(3, mode='thick'), name='e4'),
        xt.Strategy(xt.Uniform(4), name='e5')])

assert np.all(line.get_table().name == [
       'e0_entry', 'e0..entry_map', 'drift_e0..0', 'e0..0', 'drift_e0..1',
       'e0..1', 'drift_e0..2..0', 'drift_e0..2..1', 'e0..2',
       'drift_e0..3', 'e0..exit_map', 'e0_exit', 'e1_entry',
       'e1..entry_map', 'e1..0', 'e1..1_entry', 'e1..1..0', 'e1..1..1',
       'e1..1_exit', 'e1..2', 'e1..exit_map', 'e1_exit', 'e2_entry',
       'e2..entry_map', 'drift_e2..0', 'e2..0', 'drift_e2..1', 'e2..1',
       'drift_e2..2', 'e2..2', 'drift_e2..3', 'e2..exit_map', 'e2_exit',
       'drift_e3..0', 'drift_e3..1', 'drift_e3..2', 'e4_entry', 'e4..0',
       'e4..1', 'e4..2', 'e4_exit', 'e5_entry', 'drift_e5..0',
       'drift_e5..1', 'drift_e5..2', 'drift_e5..3', 'drift_e5..4',
       'e5_exit', '_end_point'])

for nn, ee in line.items():
    if nn.startswith('drift'):
        assert _is_drift(ee, line)
        assert _behaves_like_drift(ee, line)
    elif nn in ['e4..0', 'e4..1', 'e4..2']:
        assert _is_drift(ee, line)
        assert _behaves_like_drift(ee, line)
    elif nn.endswith('_entry') or nn.endswith('_exit'): # Markers
        assert not _is_drift(ee, line)
        assert _behaves_like_drift(ee, line)
    else:
        assert not _is_drift(ee, line)
        assert not _behaves_like_drift(ee, line)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json')
line.particle_ref = xt.Particles(
                    mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.build_tracker()

# Generate some particles with known normalized coordinates
particles = line.build_particles(
    nemitt_x=2.5e-6, nemitt_y=1e-6,
    x_norm=[-1, 0, 0.5], y_norm=[0.3, -0.2, 0.2],
    px_norm=[0.1, 0.2, 0.3], py_norm=[0.5, 0.6, 0.8],
    zeta=[0, 0.1, -0.1], delta=[1e-4, 0., -1e-4])

# Inspect physical coordinates
tab = particles.get_table()
tab.show()
# prints
#
# Table: 3 rows, 17 cols
# particle_id s            x           px            y          py zeta   delta chi ...
#           0 0 -0.000253245  3.33271e-06  5.10063e-05 1.00661e-06    0  0.0001   1
#           1 0 -2.06127e-09  3.32087e-07 -3.42343e-05 5.59114e-08  0.1       0   1
#           2 0  0.000152331 -7.62878e-07  3.45785e-05  1.0462e-06 -0.1 -0.0001   1


# Compute twiss
tw = line.twiss()

# Use twiss to compute normalized coordinates
norm_coord = tw.get_normalized_coordinates(particles, nemitt_x=2.5e-6,
                                           nemitt_y=1e-6)

# Inspect normalized coordinates
norm_coord.show()
#
# Table: 3 rows, 8 cols
# particle_id at_element       x_norm px_norm y_norm py_norm   zeta_norm  pzeta_norm
#           0          0           -1     0.1    0.3     0.5 1.06651e-07  0.00313799
#           1          0 -1.59607e-20     0.2   -0.2     0.6  0.00318676 1.12046e-05
#           2          0          0.5     0.3    0.2     0.8  -0.0031868  -0.0031492

#!end-doc-part

# Check that the computed normalized coordinates are correct
assert np.allclose(norm_coord['x_norm'], [-1, 0, 0.5], atol=5e-14, rtol=0)
assert np.allclose(norm_coord['y_norm'], [0.3, -0.2, 0.2], atol=5e-14, rtol=0)
assert np.allclose(norm_coord['px_norm'], [0.1, 0.2, 0.3], atol=5e-14, rtol=0)
assert np.allclose(norm_coord['py_norm'], [0.5, 0.6, 0.8], atol=5e-14, rtol=0)
